from .download import download_zip
from pathlib import Path
down_path = Path(__file__).parent
xz_url = "addons_file"+"/"+down_path.name+"/"
download_zip(xz_url + "resources.zip",down_path)
download_zip(xz_url + "weight_layers_addon.zip",down_path)


import imp
import sys

import bpy  # noqa F401

from .weight_layers_addon import get_current_level_modules
from .translation.translate import GetTranslationDict

bl_info = {
    "name": "Alpha Trees Pro 2.3.6",
    "author": "Andrew Stevenson",
    "description": "Create and use billboard textures for trees, easily",
    "blender": (3, 0, 0),
    "version": (2, 3, 6),
    "location": "3D viewport > Tool Panel > Alpha Trees",
    "wiki_url": "https://sites.google.com/view/alphatreesdocs/home",
    "tracker_url": "https://blenderartists.org/t/alpha-trees-render-massive-forests-fast/1250694?u=strike_digital",
    "support": "COMMUNITY",
    "category": "3D View",
}

modules = get_current_level_modules(__file__, __package__)

# For some reason different os's load modules in different orders, which is a problem as that can cause
# circular imports. This just sorts the modules to a predefined order no matter the operating system.
# Thanks to Ingo and Kerstin from https://braverabbit.com for figuring out the solution :)
mod_order = [
    "functions",
    "icons",
    "info",
    "operators",
    "preferences",
    "settings",
    "ui",
    "weight_layers_addon",
]
modules.sort(key=lambda m: mod_order.index(m.__name__.split(".")[-1]))

# get all modules in the addon (but only the top level weight layers module, as it reloads itsself)
if "bpy" in locals():
    at_modules = [
        sys.modules[m]
        for m in sys.modules.keys()
        if __package__ in m and "weight_layers_addon" not in m or m.endswith("weight_layers_addon")
    ]
    for m in at_modules:
        try:
            imp.reload(m)
        except ModuleNotFoundError as e:
            print(e)


def register():
    for module in modules:
        module.register()
    bpy.app.translations.register(__package__, GetTranslationDict())




def unregister():
    for module in reversed(modules):
        module.unregister()
    bpy.app.translations.unregister(__package__)
