import bpy
from bpy_extras.io_utils import ImportHelper
from bpy.props import FloatProperty, BoolProperty, StringProperty, EnumProperty

from .operator_utils import MultiTreeBase
from ..functions import imp_functions, biome_functions
from ..settings import CopyMenuItem
from ..functions import helper_functions as helpers
from ..functions.biome_functions import get_node, get_input
from ..functions.helper_functions import get_all_bsettings, get_mod_group, get_prefs, get_biome_vars,\
    get_tree_vars, get_mat_control_node, centered_row
from ..icons import get_icon, get_tree_previews, get_weight_previews
from ..weight_layers_addon.WL_functions import get_wl_obj as get_at_obj, setup_new_geo_node_tree
from ..constants import GEO_NODES_VERSION, IS_DEV, NODE_NAMES, VARIANT

import os
import json
import zipfile
import webbrowser
from os.path import join
from pathlib import Path


# Systems
class ALPHATREE_OT_update_old_biome(bpy.types.Operator):
    """Update all old biomes to the current version"""
    bl_label = "Update old biomes"
    bl_idname = "alpha_tree.update_old_biomes"
    bl_description = "Update all old biomes to the current version"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {"REGISTER", "UNDO"}

    def execute(self, context):

        old_biome_psettings = [p for p in bpy.data.particles if "AT_BIOME_" in p.name]
        new_biome_settings = {}
        preview_names = [p[0] for p in get_tree_previews()]

        obj = context.object
        bpy.ops.alpha_tree.biome_list_actions(action="ADD")
        temp_slot = obj.at.biome_list[-1]
        for psetting in old_biome_psettings[:]:
            bpy.ops.alpha_tree.sys_new_settings(add_tree=False)
            new_settings_group = temp_slot.biome_settings
            new_biome_settings[psetting.name] = new_settings_group
            new_settings = new_settings_group.at
            new_settings.name = psetting.name.replace("AT_BIOME_", "").replace("_", " ")

            control_node = get_node(new_settings_group, "control")
            # Set scale. Approximate conversion, no idea if it's actually accurate
            get_input(control_node, "scale").default_value = psetting.particle_size * (psetting.hair_length)
            # Set scale randomness
            get_input(control_node, "scale_randomness").default_value = psetting.size_random
            # Set viewport percentage
            get_input(control_node, "viewport_percentage").default_value = psetting.display_percentage
            # Set random rotation
            get_input(control_node, "tilt_randomness").default_value = psetting.rotation_factor_random

            try:
                collection = bpy.data.collections[psetting.name]
            except KeyError:
                continue

            old_tree_obj = None
            for old_tree_obj in collection.objects:
                old_material = old_tree_obj.material_slots[0].material
                # check has at material
                if old_material and old_material.name.startswith("AT_"):
                    bpy.ops.alpha_tree.tree_list_actions(action='ADD')
                    tree_list = new_settings_group.at.tree_list
                    tree = tree_list[-1]

                    # check if is just default obj
                    if "default" in old_material.name.lower():
                        continue
                    diff_leaf_node = old_material.node_tree.nodes["leaf color"]
                    # find the tree in the current library, else raise error and continue
                    if diff_leaf_node.image:
                        tree_name = diff_leaf_node.image.name.split("_")[0] + "_preview.png"
                        if tree_name in preview_names:
                            tree.selected_tree = tree_name
                        else:
                            self.report({"ERROR"}, f"Could not find tree '{tree_name.replace('_preview.png', '')}'\
                                        in the current library".replace("  ", ""))
                            continue
                    else:
                        continue

                    old_mat_control_node = old_material.node_tree.nodes["Alpha trees control"]
                    new_mat_control_node = get_mat_control_node(self, context, tree)

                    # copy material settings
                    current_group = ""
                    for old_socket in old_mat_control_node.inputs:
                        if old_socket.hide_value:
                            current_group = old_socket.name.lower()
                            continue
                        for new_socket in new_mat_control_node.inputs:
                            if new_socket.name.lower() == current_group.lower() + "_" + old_socket.name.lower():
                                new_socket.default_value = old_socket.default_value
                                break
                        else:
                            print(f"Couldn't find equivalent socket for {old_socket.name}")

                    # update tree settings
                    tree_control_node = get_node(get_mod_group(tree.tree_obj), "tree")
                    quality = diff_leaf_node.image.name.split("_")[-1]
                    if quality in ["low", "med", "high"]:
                        tree.quality = quality
                    scale = old_tree_obj.scale
                    if scale[1] == 0:
                        get_input(tree_control_node, "tree_show_render").default_value = False
                    else:
                        get_input(tree_control_node, "tree_scale").default_value = scale[2]
                        get_input(tree_control_node, "tree_width").default_value = (1 + (scale[2] - scale[0]))
                    get_input(tree_control_node, "tree_z_offset").default_value = old_tree_obj.data.vertices[0].co[0]

                else:
                    continue

                track_camera = None
                for constraint in old_tree_obj.constraints:
                    if "AT_TRACK_CONSTRAINT" in constraint.name:
                        track_camera = constraint.target
                        break
                get_input(control_node, "camera").default_value = track_camera

        # remove temp slot
        bpy.ops.alpha_tree.biome_list_actions(action="REMOVE")

        temp_collection = bpy.data.collections.new("AT_TEMP_COLLECTION")
        context.scene.collection.children.link(temp_collection)
        context.view_layer.layer_collection.children.get(temp_collection.name).hide_viewport = True
        orig_obj = context.object

        for obj in bpy.data.objects:
            psystems = [p for p in obj.particle_systems if p.settings and p.settings in old_biome_psettings]
            if psystems:
                temp_collection.objects.link(obj)
                context.view_layer.objects.active = obj

            for psystem in psystems:
                slot = obj.at.new_biome_slot(context)
                slot.biome_settings = new_biome_settings[psystem.settings.name]
                mod_group = helpers.get_mod_group(obj)
                slot_node = mod_group.nodes[str(slot.index)]
                if psystem.vertex_group_density:
                    slot.density.vgroup_enum = psystem.vertex_group_density
                get_input(slot_node, "invert_density_attr").default_value = psystem.invert_vertex_group_density
                if psystem.vertex_group_length:
                    slot.scale.vgroup_enum = psystem.vertex_group_length
                get_input(slot_node, "invert_scale_attr").default_value = psystem.invert_vertex_group_length

                pass

        context.view_layer.objects.active = orig_obj
        bpy.data.collections.remove(temp_collection)

        return {"FINISHED"}


class ALPHATREE_OT_biome_list_actions(bpy.types.Operator):
    """Manipulate list items, either: move up, move down, add or remove"""
    bl_label = "List actions"
    bl_idname = "alpha_tree.biome_list_actions"
    bl_description = "Manipulate list items, either: move up, move down, add or remove"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {"REGISTER", "UNDO"}

    action: bpy.props.EnumProperty(
        items=(
            ("ADD", "Add", "Add item"),
            ("REMOVE", "Remove", "Remove item"),
            ("UP", "Up", "Move item up"),
            ("DOWN", "Down", "Move item down"),
        ),
        options={"HIDDEN"},
    )

    def execute(self, context):
        biome_list, list_settings, b_index, _, _ = get_biome_vars(self, context)
        obj = get_at_obj(context)
        mod_group = helpers.get_mod_group(obj)
        if mod_group:
            nodes = mod_group.nodes
            slot_nodes = [n for n in nodes if n.type == "DISTRIBUTE_POINTS_ON_FACES" or n.type == "GROUP"]

        if self.action == "ADD":

            obj.at.new_biome_slot(context)
            return {"FINISHED"}

        elif self.action == "REMOVE":

            if len(biome_list) == 1:
                bpy.data.node_groups.remove(mod_group)
                mod = obj.modifiers.get("AT_GEO_NODES")
                obj.modifiers.remove(mod)
            else:
                for n in slot_nodes:
                    if int(n.name) == b_index:
                        slot_nodes.remove(n)
                        mod_group.nodes.remove(n)
                        break

                def node_sort(node):
                    return int(node.name)

                slot_nodes.sort(key=node_sort)

                for n in slot_nodes:
                    if int(n.name) > b_index:
                        n.name = str(int(n.name) - 1)

                biome_functions.organise_list_group(mod_group)

            biome_list.remove(b_index)
            list_settings.b_index = min(max(0, b_index - 1), len(biome_list) - 1)

            return {"FINISHED"}

        elif self.action == "UP" or "DOWN":
            up = (self.action == "UP")
            length = len(biome_list) - 1
            new_index = b_index + (-1 if up else 1)
            new_index = max(0, min(new_index, length))

            nodes = mod_group.nodes
            indices = {}
            for n in slot_nodes:
                indices[n.name] = n

            indices[str(b_index)].name = str(new_index)
            indices[str(new_index)].name = str(b_index)
            indices[str(b_index)].name = str(new_index)

            biome_list.move(new_index, b_index)
            list_settings.b_index = max(0, min(new_index, length))

        return {"FINISHED"}


class ALPHATREE_OT_save_biome(bpy.types.Operator):
    """Save the current biomes"""
    bl_label = "Save biome settings"
    bl_idname = "alpha_tree.save_biome"
    bl_description = "Save the current biomes as an external file that can be accessed from any blend file"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {"REGISTER"}

    name: StringProperty(name="Name", default="My biome")

    @classmethod
    def poll(cls, context):
        biome = get_biome_vars(None, context)[-1]
        if biome:
            if biome.biome_settings != "NONE":
                return True
        return False

    def invoke(self, context, event):
        biome_list, _, _, _, biome = get_biome_vars(self, context)

        at = context.scene.alpha_trees
        copy_items = at.copy_items
        copy_items.clear()

        for biome in biome_list:
            if biome.biome_settings:
                copy_items.add()
                new_item = copy_items[len(copy_items) - 1]
                new_item.on = False

                new_item.name = biome.biome_settings.at.name
                new_item.string = biome.biome_settings.name

        return context.window_manager.invoke_props_dialog(self)

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True
        layout.prop(self, "name")

        biome_path = join(get_prefs().lib_path, "biomes")
        fname = self.name.replace(" ", "-").lower().replace("_", "-")
        fname = fname + ".json"

        if os.path.isfile(join(biome_path, fname)):
            row = layout.row()
            row.alert = True
            row.label(text="This biome already exists, will be overwritten")

        at = context.scene.alpha_trees
        copy_items = at.copy_items
        box = layout.box()
        row = box.row()
        row.alignment = "CENTER"
        row.label(text="Save biomes:")

        col = box.column(align=True)
        col.use_property_split = False
        for i in copy_items:
            col.prop(i, "on", text=i.name, toggle=True, icon="CHECKMARK" if i.on else "BLANK1")

    def execute(self, context):
        lib_path = get_prefs().lib_path
        biome_list, _, _, _, biome = get_biome_vars(self, context)
        at = context.scene.alpha_trees

        copy_items = at.copy_items

        # check if any selected
        for i in copy_items:
            if i.on:
                break
        else:
            return {"FINISHED"}

        if "biomes" in os.listdir(lib_path):
            bf = join(lib_path, "biomes")
        else:
            bf = os.mkdir(join(lib_path, "biomes"))

        biomes = []
        for i in copy_items:
            if i.on:
                for biome in biome_list:
                    if biome.biome_settings.name == i.string:
                        biomes.append(biome)

        data = biome_functions.create_biome_json(self, context, biomes)

        fname = self.name.replace(" ", "-").lower().replace("_", "-")
        with open(join(bf, fname + ".json"), 'w') as outfile:
            json.dump(data, outfile, indent=4)

        self.report({"INFO"}, f"Saved {fname + '.json'}")

        return {"FINISHED"}


class ALPHATREE_OT_load_biome(bpy.types.Operator):
    """Load previously saved biome settings"""
    bl_label = "Load biome settings"
    bl_idname = "alpha_tree.load_biome"
    bl_description = "Load previously saved biome settings"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {"REGISTER"}

    # load_weights: BoolProperty(name="Load weights",
    #                            description="Generate the procedural weight maps from the original biome\
    #         (This slow down importing depending on)"                                                    ,
    #                            default=True)

    old_biomes = []

    def invoke(self, context, event):
        biome_path = join(get_prefs().lib_path, "biomes")
        at = context.scene.alpha_trees
        copy_items = at.copy_items
        copy_items.clear()
        self.old_biomes = []

        for f in os.listdir(biome_path):
            if "_collection_" in f or "_single_" in f:
                # ignore old biomes
                self.old_biomes.append(f)
                continue

            copy_items.add()
            new_item = copy_items[len(copy_items) - 1]
            new_item.name = f.capitalize().replace("-", " ").removesuffix(".json")
            new_item.string = f
            new_item.on = False

        return context.window_manager.invoke_props_dialog(self, width=300)

    def draw(self, context):
        layout = self.layout
        at = context.scene.alpha_trees
        copy_items = at.copy_items

        row = layout.row()
        col1 = row.column(align=True)
        box = col1.box().column(align=True)
        box.label(text="Import biomes:", icon_value=get_icon("biome.png"))
        box = col1.box().column(align=True)

        files = 0

        for item in copy_items:
            files += 1
            box.prop(item, "on", text=item.name, toggle=True, icon="CHECKMARK" if item.on else "BLANK1")

        if self.old_biomes:
            box = col1.box().column(align=True)
            box.enabled = False
            row = box.row(align=True)
            row.enabled = True
            row.label(text="Biomes from previous versions cannot be imported:", icon="ERROR")

        for old_biome in self.old_biomes:
            files += 1
            text = old_biome.split("_")[0].capitalize().replace("-", " ")
            box.operator(self.bl_idname, text=text, icon="BLANK1")

        if not files:
            box.label(text="No biomes to import")

    def execute(self, context):
        biome_path = join(get_prefs().lib_path, "biomes")
        at = context.scene.alpha_trees
        copy_items = at.copy_items
        not_found = []

        for item in copy_items:
            if item.on:
                with open(join(biome_path, item.string)) as file:
                    data = json.load(file)
                    i = biome_functions.load_biome_json(self, context, data)
                    not_found.extend(i)

        if not_found:
            for nf in not_found:
                self.report({"ERROR"}, f"Could not find {nf}")

        return {"FINISHED"}


class ALPHATREE_OT_remove_biome(bpy.types.Operator):
    """Remove previously saved biomes"""
    bl_label = "Remove saved biome"
    bl_idname = "alpha_tree.remove_saved_biome"
    bl_description = "Remove a previously saved biome"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {"REGISTER"}

    show_warning: BoolProperty(name="Show warning", default=True)

    display_name: StringProperty()

    file_name: StringProperty()

    def invoke(self, context, event):
        if self.show_warning:
            return context.window_manager.invoke_props_dialog(self)
        else:
            return self.execute(context)

    def draw(self, context):
        layout: bpy.types.UILayout = self.layout
        col1 = layout.column(align=True)
        column = col1.box().column(align=True)
        row = column.row(align=True)
        row.alignment = "CENTER"
        row.label(text="Delete biome: " + self.display_name)

        column = col1.box().column(align=True)
        column.alert = True
        row = column.row(align=True)
        row.alignment = "CENTER"
        row.label(text="Warning: you cannot undo this action")
        row = column.row(align=True)
        row.alignment = "CENTER"
        row.label(text="Are you sure you want to continue?")

    def execute(self, context):
        biome_path = Path(get_prefs().lib_path) / "biomes"
        biome_file = biome_path / self.file_name

        if not biome_file.exists():
            self.report({"ERROR"}, f"{str(biome_file)} does not exist")
        elif not biome_file.name.endswith(".json"):
            self.report({"ERROR"}, f"{str(biome_file)} is not a json file")

        try:
            os.remove(str(biome_file))
        except Exception as e:
            self.report({"ERROR"}, f"Could not remove {str(biome_file)}:\n{e}")

        return {"FINISHED"}


class ALPHATREE_OT_manage_biomes(bpy.types.Operator):
    """Manage previously saved biomes"""
    bl_label = "Manage saved biomes"
    bl_idname = "alpha_tree.manage_saved_biomes"
    bl_description = "Manage previously saved biomes"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {"REGISTER"}

    contents = {}

    def invoke(self, context, event):
        biome_path = join(get_prefs().lib_path, "biomes")
        at = context.scene.alpha_trees
        copy_items: list[CopyMenuItem] = at.copy_items
        copy_items.clear()
        self.old_biomes = []

        for f in os.listdir(biome_path):
            if not f.endswith(".json"):
                continue

            if "_collection_" in f or "_single_" in f:
                # ignore old biomes
                self.old_biomes.append(f)
                continue

            copy_items.add()
            new_item = copy_items[len(copy_items) - 1]
            new_item.name = f.capitalize().replace("-", " ").removesuffix(".json")
            new_item.string = f
            new_item.on = False

            with open(join(biome_path, f)) as file:
                data = json.load(file)

            self.contents[f] = data

        return context.window_manager.invoke_props_dialog(self, width=310)

    def draw(self, context):
        layout = self.layout
        at = context.scene.alpha_trees
        copy_items = at.copy_items
        biome_path: Path = Path(get_prefs().lib_path) / "biomes"

        row = layout.row()
        col1 = row.column(align=True)
        box = col1.box().column(align=False)
        box.label(text="Biomes:")
        box = col1.box().column(align=False)
        scale = 1.5
        box.scale_y = scale
        box.scale_x = scale

        files = 0

        for item in copy_items:
            if not (biome_path / item.string).exists():
                continue
            sub = box.column(align=True)
            files += 1
            row = sub.row(align=True)
            row.prop(item, "on", text="", icon="TRIA_DOWN" if item.on else "TRIA_RIGHT")
            row.prop(item, "name", text="")
            op = row.operator("alpha_tree.remove_saved_biome", text="", icon="TRASH")
            op.display_name = item.name
            op.file_name = item.string
            if item.on:
                sub = sub.box().column(align=True)
                sub.scale_y = 1 / scale
                sub.enabled = False
                data: dict = self.contents[item.string]
                addon_version = ".".join([str(i) for i in data.get("addon_version", (2, 0, 0))])
                sub.label(text="Addon version: " + addon_version)
                biome_version = ".".join([str(i) for i in data.get("version", (1, 0, 0))])
                sub.label(text="Biome version: " + biome_version)

                sub.label(text="Biomes:")
                for name, biome in data["biomes"].items():
                    sub.label(text="    " + name)
                    for name in biome["trees"].keys():
                        sub.label(text="        " + name)

        if self.old_biomes:
            box = col1.box().column(align=False)
            box.scale_y = scale
            box.scale_x = scale
            row = box.row(align=True)
            row.label(text="Biomes from previous versions cannot be imported:", icon="ERROR")

        for old_biome in self.old_biomes:
            if not (biome_path / old_biome).exists():
                continue
            files += 1
            text = old_biome.split("_")[0].capitalize().replace("-", " ")
            row = box.row(align=True)
            box1 = row.box()
            box1.scale_y = 1 / scale
            box1.label(text=text)
            op = row.operator("alpha_tree.remove_saved_biome", text="", icon="TRASH", depress=True)
            op.display_name = text
            op.file_name = old_biome
            op.show_warning = False

        if not files:
            box.label(text="No biomes to import")

    def execute(self, context):
        biome_path = Path(get_prefs().lib_path) / "biomes"
        at = context.scene.alpha_trees
        copy_items: CopyMenuItem = at.copy_items
        not_found = []
        renamed = 0

        for item in copy_items:
            file_path: Path = biome_path / item.string
            name: str = item.name
            new_name = name.lower().replace(" ", "-") + ".json"
            if new_name != item.string:
                if not file_path.exists():
                    not_found.append(file_path)
                    continue
                new_path = file_path.parent / new_name
                try:
                    os.rename(file_path, new_path)
                except Exception as e:
                    self.report("ERROR", f"Could not rename {file_path} to {new_path}: \n{e}")
                renamed += 1

        if not_found:
            for nf in not_found:
                self.report({"ERROR"}, f"Could not find {str(nf)}")

        if renamed:
            suffix = "s" if renamed != 1 else ""
            self.report({"INFO"}, f"Renamed {renamed} biome{suffix}")
        return {"FINISHED"}


class ALPHATREE_OT_export_package(MultiTreeBase, bpy.types.Operator):
    """Export a package as a zip file that can be installed by another user"""
    bl_label = "Export package"
    bl_idname = "alpha_tree.export_package"
    bl_description = "Export a package as a zip file that can be installed by another user"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {"REGISTER"}

    show_license: BoolProperty(
        name="Show license info",
        default=False,
    )

    tscale: FloatProperty(name="Scale", default=6.0, min=3, max=15)

    tsearch: StringProperty(name="Search", description="Filter trees by keywords", default="")

    show_biomes: BoolProperty(
        name="Show biomes",
        default=True,
    )

    bscale: FloatProperty(name="Scale", default=1.5, min=0.75, max=5)

    bsearch: StringProperty(name="Search", description="Filter biomes by keywords", default="")

    show_settings: BoolProperty(
        name="Show settings",
        default=True,
    )

    name: StringProperty(name="Name", description="Name of package", default="Custom package")

    open_path: BoolProperty(
        name="Open path when done",
        default=True,
    )

    pack_biome_trees: BoolProperty(
        name="Pack trees specified in biomes",
        description="Pack the trees that are in the selected biomes, even if they are not selected\
            (recommended if others are going to use it)",
        default=True,
    )

    def invoke(self, context, event):

        at = context.scene.alpha_trees
        copy_items = at.copy_items
        copy_items.clear()

        lib_path = get_prefs().lib_path
        tree_path = join(lib_path, "rendered_maps")

        for img in os.listdir(tree_path):
            if "preview.png" in img:
                copy_items.add()
                new_item = copy_items[len(copy_items) - 1]
                new_item.on = False
                tree_name = img.split("_")[0]
                new_item.name = tree_name
                new_item.string = img

        biome_path = join(lib_path, "biomes")

        for biome in os.listdir(biome_path):
            copy_items.add()
            new_item = copy_items[len(copy_items) - 1]
            new_item.on = False
            new_item.name = biome.split("_")[0].replace("-", " ")

            new_item.string = biome

        try:
            _ = get_icon("default.png", type="preview")
        except KeyError:
            imp_functions.get_previews_from_files(self, context)

        return context.window_manager.invoke_props_dialog(self, width=1000)

    def draw(self, context):
        layout = self.layout
        at = context.scene.alpha_trees
        copy_items = at.copy_items

        tbox = layout.box()
        trow = tbox.row()
        trow.prop(self, "show_trees", text="", icon="TRIA_DOWN" if self.show_trees else "TRIA_RIGHT")

        if self.show_trees:
            row = tbox.row()
            row.prop(self, "tsearch", text="", icon="VIEWZOOM")
            row.prop(self, "tscale")
            row = row.row(align=True)
            op = row.operator("alpha_tree.toggle_copy_items", text="", icon_value=get_icon("toggle.png"))
            op.key = ".png," + self.tsearch
            op.type = "TOGGLE"
            op = row.operator("alpha_tree.toggle_copy_items", text="", icon_value=get_icon("invert.png"))
            op.key = ".png," + self.tsearch
            op.type = "INVERT"

        bbox = layout.box()
        brow = bbox.row()
        brow.prop(self, "show_biomes", text="", icon="TRIA_DOWN" if self.show_biomes else "TRIA_RIGHT")

        if self.show_biomes:
            row = bbox.row()
            row.prop(self, "bsearch", text="", icon="VIEWZOOM")
            row.prop(self, "bscale")

        if ".png" in copy_items[0].string and self.show_trees:
            tgrid = tbox.grid_flow(row_major=True, columns=0)
        elif self.show_trees:
            tbox.label(text="No trees yet in library")

        if ".json" in copy_items[-1].string and self.show_biomes:
            split = bbox.split()
            box1 = split.box()
            box1.scale_y = self.bscale
            row = box1.row()
            row.label(text="Single:", icon_value=get_icon("biome.png"))
            op = row.operator("alpha_tree.toggle_copy_items", emboss=False, text="", icon_value=get_icon("toggle.png"))
            op.key = ".json,single," + self.bsearch
            op.type = "TOGGLE"
            op = row.operator("alpha_tree.toggle_copy_items", emboss=False, text="", icon_value=get_icon("invert.png"))
            op.key = ".json,single," + self.bsearch
            op.type = "INVERT"

            box2 = split.box()
            box2.scale_y = self.bscale
            row = box2.row()
            row.label(text="Collection:", icon="GROUP")
            op = row.operator("alpha_tree.toggle_copy_items", emboss=False, text="", icon_value=get_icon("toggle.png"))
            op.key = ".json,collection," + self.bsearch
            op.type = "TOGGLE"
            op = row.operator("alpha_tree.toggle_copy_items", emboss=False, text="", icon_value=get_icon("invert.png"))
            op.key = ".json,collection," + self.bsearch
            op.type = "INVERT"

        elif self.show_biomes:
            bbox.label(text="No biomes yet in library")

        tree = False
        tree_num = 0
        single = False
        coll = False
        biome_num = 0

        for i in copy_items:
            if ".png" in i.string:
                if i.on:
                    tree_num += 1
                if self.show_trees:
                    if self.tsearch.lower() in i.string.lower() or not self.tsearch:
                        tree = True
                        box = tgrid.box()
                        col = box.column()
                        i_val = get_icon(i.string, type="preview")
                        col.template_icon(icon_value=i_val, scale=self.tscale)
                        col.scale_y = self.tscale / 5
                        col.prop(i, "on", text=i.name, toggle=True, icon="CHECKMARK" if i.on else "BLANK1")

            else:
                if i.on:
                    biome_num += 1
                if self.bsearch.lower() in i.string.lower() or not self.bsearch:
                    if ".json" in copy_items[-1].string and self.show_biomes:
                        type = i.string.split("_")[1]
                        if "single" in type:
                            box = box1
                            single = True
                        elif "collection" in type:
                            box = box2
                            coll = True
                        else:
                            print("invalid biome")
                            continue

                        box.prop(i, "on", text=i.name, toggle=True, icon="CHECKMARK" if i.on else "BLANK1")

        if not tree and self.show_trees:
            tbox.label(text="No trees with that name")
        if not single and self.show_biomes:
            box1.label(text="No single biomes with that name" if self.bsearch else "No single biomes")
        if not coll and self.show_biomes:
            box2.label(text="No biome collections with that name" if self.bsearch else "No biome collections")

        tstr_num = str(tree_num) if tree_num else ""
        bstr_num = str(biome_num) if biome_num else ""
        trow.label(text="Trees: " + tstr_num)
        brow.label(text="Biomes: " + bstr_num)

        # settings
        box = layout.box()
        row = box.row()
        row.prop(self, "show_settings", text="", icon="TRIA_DOWN" if self.show_settings else "TRIA_RIGHT")
        row.label(text="Settings:")

        if self.show_settings:
            box.prop(self, "name")

            for f in os.listdir(bpy.path.abspath(get_prefs().export_dir)):
                if self.name.replace(" ", "-").replace("_", "-").replace(".zip", "") + ".zip" in f:
                    row = box.row()
                    row.alert = True
                    row.label(text="This file already exists and will be overwritten")

            box.prop(self, "open_path")
            box.prop(self, "pack_biome_trees")
            box.prop(self, "show_license", icon="INFO")

            if self.show_license:
                col = box.column(align=True)
                col.active = True
                col.scale_y = 0.7

                text = [
                    "If you are the creator of the original 3D models used to make these alpha trees",
                    "or they have a CC0 license or similar, then you are free to use the exported package",
                    "in any way you like, commercial or otherwise.", "",
                    "However, if you are not the original creator, you will need express permission from them",
                    "if you want to distribute the trees to other people."
                ]

                imp_functions.draw_text(col, text)

    def execute(self, context):

        prefs = get_prefs()
        lib_path = prefs.lib_path
        tree_path = join(lib_path, "rendered_maps")
        biome_path = join(lib_path, "biomes")
        exp_dir = bpy.path.abspath(prefs.export_dir)  # pylint: disable=assignment-from-no-return
        z_name = self.name.replace(" ", "-").replace("_", "-").replace(".zip", "") + ".zip"
        exp_file = join(exp_dir, z_name)
        at = context.scene.alpha_trees
        copy_items = at.copy_items
        not_found = []
        added = []

        self.report({"INFO"}, "Exporting package...")

        if not os.path.isdir(biome_path):
            self.report({"ERROR"}, "No directory found for biomes")

        with zipfile.ZipFile(exp_file, "w") as zip:
            for i in copy_items:
                if i.on:
                    if ".json" in i.string:
                        for biome in os.listdir(biome_path):
                            if i.string in biome:
                                fpath = join(biome_path, biome)
                                zip.write(fpath, biome)

                                if self.pack_biome_trees:
                                    type = biome.split("_")[1]
                                    with open(fpath) as contents:
                                        data = json.load(contents)
                                        if type == "collection":
                                            biomes = data["biomes"]
                                        else:
                                            biomes = [data]

                                        for biome in biomes:
                                            trees = biome["trees"]
                                            for tree in trees:
                                                if tree["selected_tree"] not in added:
                                                    found = False
                                                    for img in os.listdir(tree_path):
                                                        if tree["name"] in img:
                                                            fpath = join(tree_path, img)
                                                            try:
                                                                zip.write(fpath, img)
                                                            except UserWarning:
                                                                pass
                                                            found = True

                                                    if not found:
                                                        not_found.append(tree["name"])
                                                added.append(tree["selected_tree"])

                    else:
                        if i.string not in added:
                            found = False
                            for img in os.listdir(tree_path):
                                if i.name in img:
                                    fpath = join(tree_path, img)
                                    zip.write(fpath, img)
                                    found = True
                            if not found:
                                not_found.append(tree["name"])
                            added.append(i.string)

        for nf in not_found:
            self.report({"ERROR"}, f"Could not find {nf}")

        if self.open_path:
            webbrowser.open('file:///' + exp_dir)

        self.report({"INFO"}, "Successfully exported package")

        return {'FINISHED'}


class ALPHATREE_OT_install_package(bpy.types.Operator, ImportHelper):
    """Install an external package as a zip file into the local library"""
    bl_label = "Install package"
    bl_idname = "alpha_tree.install_package"
    bl_description = "Install an external package as a zip file into the local library"
    bl_options = {"REGISTER"}

    filename_ext = ".zip"

    filter_glob: StringProperty(
        default="*.zip",
        options={'HIDDEN'},
        maxlen=255,  # Max internal buffer length, longer would be clamped.
    )

    overwrite: BoolProperty(name="Overwrite",
                            description="Overwrite files if they already exist in the library",
                            default=False)

    def execute(self, context):

        zipf = self.filepath  # pylint: disable=no-member
        lib_path = get_prefs().lib_path

        if not os.path.isfile(zipf):
            self.report({"ERROR"}, "No zip file selected")

        try:
            with zipfile.ZipFile(zipf) as _:
                pass
        except zipfile.BadZipFile:
            self.report({"ERROR"}, "Can't install, zip file is corrupted")

        with zipfile.ZipFile(zipf, mode='r', allowZip64=True) as zip:
            for f in zip.namelist():
                if ".json" in f:
                    path = join(lib_path, "biomes")
                else:
                    path = join(lib_path, "rendered_maps")

                if os.path.isfile(join(path, f)) and not self.overwrite:
                    print(f"file already exists: {f}, skipping")
                    continue

                zip.extract(f, path)

        bpy.ops.alpha_tree.reload_previews()  # pylint: disable=no-member
        self.report({"INFO"}, "Installed package: " + zipf)

        return {'FINISHED'}


class ALPHATREE_OT_new_settings(bpy.types.Operator):
    """Add new settngs for this system"""
    bl_label = "Add settings"
    bl_idname = "alpha_tree.sys_new_settings"
    bl_description = "Add new settngs for this system"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {"REGISTER", "UNDO"}

    add_tree: BoolProperty(default=True)

    def execute(self, context):

        biome_list, _, b_index, old_slot_node, biome = get_biome_vars(self, context)
        obj = get_at_obj(context)
        orig_obj = context.object

        modifier_group = helpers.get_mod_group(obj)
        settings_group = biome_functions.get_at_node_group(NODE_NAMES["slot"], type="BIOME")

        nodes = modifier_group.nodes
        links = modifier_group.links

        if old_slot_node.type == "DISTRIBUTE_POINTS_ON_FACES":
            old_density = old_slot_node.inputs["Density Max"].default_value
            old_scale = float(old_slot_node.label)
        else:
            old_control_node = get_node(old_slot_node.node_tree, "control")
            old_density = get_input(old_control_node, "density").default_value
            old_scale = get_input(old_control_node, "scale").default_value

            slot_node = nodes.new("GeometryNodeGroup")
            slot_node.node_tree = settings_group
            get_input(slot_node, "seed").default_value = b_index

            from_socket = old_slot_node.inputs[0].links[0].from_socket
            to_socket = old_slot_node.outputs[0].links[0].to_socket
            nodes.remove(old_slot_node)
            slot_node.name = str(b_index)
            links.new(from_socket, slot_node.inputs[0])
            links.new(slot_node.outputs[0], to_socket)

        biome_functions.organise_list_group(modifier_group)

        # collections
        collections = [col.name for col in context.scene.collection.children]
        parent_coll_name = "ALPHA_TREES_PARTICLE_COLLECTION"

        if parent_coll_name in collections:
            parent_coll = context.scene.collection.children[parent_coll_name]
        else:
            parent_coll = bpy.data.collections.new(parent_coll_name)
            try:
                # only in 2.91 and later
                parent_coll.color_tag = "COLOR_04"
            except AttributeError:
                pass
            context.scene.collection.children.link(parent_coll)
            layer_parent_coll = context.view_layer.layer_collection.children.get(parent_coll_name)
            layer_parent_coll.exclude = True

        index = 0
        while True:
            try:
                name = NODE_NAMES["slot"] + "_Biome settings " + str(index)
                _ = bpy.data.node_groups[name]
                index += 1
            except KeyError:
                break

        settings_group.name = name
        settings_group.at.type = "SETTINGS"
        settings_group.at.version = GEO_NODES_VERSION
        biome.biome_settings = settings_group
        control_node = get_node(settings_group, "control")

        collection = bpy.data.collections.new(name)
        get_input(control_node, "collection").default_value = collection
        settings_group.at.name = "Biome settings " + str(index)

        # density = biome_functions.get_obj_density(context.object)
        get_input(control_node, "density").default_value = old_density
        get_input(control_node, "scale").default_value = old_scale

        try:
            # only in 2.91 and later
            collection.color_tag = "COLOR_08"
        except AttributeError:
            pass
        parent_coll.children.link(collection)

        # placeholder ob
        default_plane = helpers.get_default_obj(self, context, control_node)

        collection.objects.link(default_plane)
        context.view_layer.objects.active = orig_obj

        if get_prefs().increase_bounces and context.scene.render.engine == "CYCLES":
            if context.scene.cycles.transparent_max_bounces < 35:
                context.scene.cycles.transparent_max_bounces = 35

        if self.add_tree:
            bpy.ops.alpha_tree.tree_list_actions(action='ADD')

        return {'FINISHED'}


class ALPHATREE_OT_remove_settings(bpy.types.Operator):
    """Remove selected setting for this system"""
    bl_label = "Remove settings"
    bl_idname = "alpha_tree.sys_remove_settings"
    bl_description = "Remove selected setting"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {"REGISTER"}

    def execute(self, context):

        _, _, b_index, _, biome = get_biome_vars(self, context)

        obj = get_at_obj(context)
        mod_group = helpers.get_mod_group(obj)
        node = mod_group.nodes[str(b_index)]
        settings_group = node.node_tree
        control_node = get_node(settings_group, "control")
        collection = get_input(control_node, "collection").default_value

        for object in bpy.data.objects:
            for list_item in object.at.biome_list:
                if list_item.biome_settings == biome.biome_settings:
                    list_item.biome_settings = None

        for obj in collection.objects:
            if obj.material_slots:
                mat = obj.material_slots[0].material
                for node in mat.node_tree.nodes:
                    if node.type == "TEX_IMAGE":
                        if node.image:
                            if node.image.users == 1:
                                bpy.data.images.remove(node.image)
                if mat.users == 1:
                    bpy.data.materials.remove(mat)
                bpy.data.meshes.remove(obj.data)
        bpy.data.collections.remove(collection)
        bpy.data.node_groups.remove(settings_group)

        return {'FINISHED'}


class ALPHATREE_OT_new_vert_group(bpy.types.Operator):
    """Add a new vertex group"""
    bl_label = "Add group"
    bl_idname = "alpha_tree.new_vert_group"
    bl_description = "Add a new vertex group"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {"REGISTER", "UNDO"}

    density: BoolProperty()

    vg_type: EnumProperty(items=get_weight_previews, name="Type")

    def invoke(self, context, event):
        return context.window_manager.invoke_props_dialog(self)

    def draw(self, context):
        if get_at_obj(context).type != "MESH":
            return

        # biome = get_biome_vars(self, context)[-1]
        # new = biome.dist.density_vert_group == "NONE" if self.density else biome.dist.scale_vert_group == "NONE"
        layout = self.layout
        col = layout.column()
        # col.use_property_split = True

        col.label(text="Type:")
        box = col.box()
        box.template_icon_view(self, "type", show_labels=True)
        row = box.row()
        row.alignment = "CENTER"
        row.label(text=self.vg_type.capitalize())

    def execute(self, context):
        # obj = get_at_obj(context)
        # at = obj.at
        # groups = obj.vertex_groups
        # biome = get_biome_vars(self, context)[-1]
        # group_name = biome.dist.density_vert_group if self.density else biome.dist.scale_vert_group
        # new = group_name == "NONE" or not self.as_layer

        # def update_history(group):
        #     # update history
        #     dict = eval(at.vg_history)
        #     if new or self.blend == "SET":
        #         dict[group.name] = [[self.vg_type, self.blend, self.gain, self.offset, self.invert]]
        #     else:
        #         dict[group_name].append([self.vg_type, self.blend, self.gain, self.offset, self.invert])
        #     at.vg_history = str(dict)

        # if self.density:
        #     name = "Density-" + self.vg_type.capitalize() + " " + str(len(groups))
        #     group = groups.new(name=name)

        #     update_history(group)

        #     if new:
        #         biome.dist.density_vert_group = group.name
        # else:
        #     name = "Scale-" + self.vg_type.capitalize() + " " + str(len(groups))
        #     group = groups.new(name=name)

        #     update_history(group)

        #     if new:
        #         biome.dist.scale_vert_group = group.name

        # if self.vg_type != "BLANK":
        #     sys_functions.create_weight_map(self, context, group)

        # if self.invert:
        #     bpy.ops.object.vertex_group_invert()

        # bpy.ops.object.vertex_group_levels(offset=self.offset, gain=self.gain)
        # groups.active = group
        # bpy.ops.object.mode_set(mode='WEIGHT_PAINT')

        # if not new:
        #     mod = obj.modifiers.new("Mix weights", "VERTEX_WEIGHT_MIX")
        #     mod.vertex_group_a = group_name
        #     mod.vertex_group_b = group.name
        #     mod.mix_mode = self.blend
        #     mod.mix_set = "ALL"
        #     bpy.ops.object.modifier_apply(modifier=mod.name)

        #     obj.vertex_groups.remove(group)
        #     groups.active = obj.vertex_groups[group_name]

        # sys_functions.check_vg_history(self, context)

        return {'FINISHED'}


class ALPHATREE_OT_remove_vert_group(bpy.types.Operator):
    """Remove a vertex group"""
    bl_label = "Remove group"
    bl_idname = "alpha_tree.remove_vert_group"
    bl_description = "Remove a vertex group"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {"REGISTER", "UNDO"}

    density: BoolProperty(options={"HIDDEN"})

    def execute(self, context):

        return {'FINISHED'}


class ALPHATREE_OT_edit_vert_group(bpy.types.Operator):
    """Edit current vertex group"""
    bl_label = "Edit group"
    bl_idname = "alpha_tree.edit_vert_group"
    bl_description = "Edit current vertex group"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {"REGISTER"}

    density: BoolProperty()

    def execute(self, context):
        obj = get_at_obj(context)
        groups = obj.vertex_groups
        biome = get_biome_vars(self, context)[-1]

        if self.density:
            group = groups[biome.dist.density_vert_group]
        else:
            group = groups[biome.dist.scale_vert_group]

        painting = obj.mode == "WEIGHT_PAINT"
        if groups.active != group and painting:
            groups.active = group
        else:
            groups.active = group
            if painting:
                bpy.ops.object.mode_set(mode='OBJECT')
            else:
                bpy.ops.object.mode_set(mode='WEIGHT_PAINT')

        return {'FINISHED'}


class ALPHATREE_OT_vert_group_actions(bpy.types.Operator):
    """Manipulate vertex groups, either: add, remove or edit them"""
    bl_label = "Vertex group actions"
    bl_idname = "alpha_tree.vert_group_actions"
    bl_description = "Manipulate vertex groups, either: add, remove or edit them"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {"REGISTER", "UNDO"}

    density: BoolProperty(options={"HIDDEN"})

    action: EnumProperty(
        items=(
            ("ADD", "Add", "A a verted group"),
            ("REMOVE", "Remove", "Remove this vertex group"),
            ("EDIT", "Edit", "Edit this vertex group"),
        ),
        options={"HIDDEN"},
    )

    vg_type: EnumProperty(items=get_weight_previews, name="Type")

    def invoke(self, context, event):
        if self.action == "ADD":
            return context.window_manager.invoke_props_dialog(self)
        else:
            return self.execute(context)

    def draw(self, context):
        col = self.layout.column()
        col.label(text="Type:")
        box = col.box()
        box.template_icon_view(self, "vg_type", show_labels=True)
        row = box.row()
        row.alignment = "CENTER"
        row.label(text=self.vg_type.capitalize())

    def execute(self, context):
        biome = get_biome_vars(self, context)[-1]
        dist = biome.density if self.density else biome.scale
        obj = dist.id_data

        if self.action == "ADD":
            vg_type = self.vg_type.removeprefix("weight_").removesuffix(".png")
            vgroup = dist.new_vgroup(context, type=vg_type)
            bpy.ops.object.mode_set(mode="WEIGHT_PAINT")

        elif self.action == "REMOVE":
            dist.remove_vgroup(context)

        elif self.action == "EDIT":
            vgroup = dist.vgroup

            mode = "WEIGHT_PAINT" if obj.mode == "OBJECT" or obj.vertex_groups.active != vgroup else "OBJECT"
            bpy.ops.object.mode_set(mode=mode)
            obj.vertex_groups.active = vgroup

        return {"FINISHED"}


# Trees


class ALPHATREE_OT_tree_list_actions(bpy.types.Operator):
    """Manipulate list items, either: move up, move down, add or remove"""
    bl_label = "List actions"
    bl_idname = "alpha_tree.tree_list_actions"
    bl_description = "Manipulate list items, either: move up, move down, add or remove"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {"REGISTER", "UNDO"}

    action: bpy.props.EnumProperty(
        items=(
            ("ADD", "Add", "Add item"),
            ("REMOVE", "Remove", "Remove item"),
            ("UP", "Up", "Move item up"),
            ("DOWN", "Down", "Move item down"),
        ),
        options={"HIDDEN"},
    )

    settings_group: bpy.props.StringProperty(
        default="",
        options={"HIDDEN"},
    )

    @staticmethod
    def add_tree(tree_list, settings_group):
        tree_list.add()
        new_tree = tree_list[len(tree_list) - 1]
        new_tree.name = "Tree " + str(len(tree_list) - 1)
        settings_group.at.t_index = len(tree_list) - 1
        control_node = get_node(settings_group, "control")
        collection = get_input(control_node, "collection").default_value

        if len(tree_list) != 1:
            # calls selected tree update function
            new_tree.selected_tree = "default.png"
            if tree_list[0].quality:
                new_tree.quality = tree_list[0].quality
            prev_tree = tree_list[-2]
            new_tree.tree_search = prev_tree.tree_search
            new_tree.tree_type = prev_tree.tree_type
            new_tree.tree_subtype = prev_tree.tree_subtype
            new_tree.tree_tag = prev_tree.tree_tag
            new_tree.show_tree_search_settings = prev_tree.show_tree_search_settings
        else:
            new_tree.tree_obj = collection.objects[0]
            new_tree.quality = "med"

        # get tree node group
        tree_obj = new_tree.tree_obj
        node_modifier = helpers.get_mod_group(tree_obj, modifier=True)
        if not node_modifier:
            node_modifier = tree_obj.modifiers.new("AT_GEO_NODE_TREE", type="NODES")
            mod_group = node_modifier.node_group
            # In 3.2 adding a new modifier doesn't automatically add a node group
            if not mod_group:
                mod_group = setup_new_geo_node_tree(name="temp")
                node_modifier.node_group = mod_group
            mod_group.name = "AT_GEO_NODE_TREE_" + tree_obj.name
        else:
            mod_group = node_modifier.node_group

        try:
            _ = get_node(mod_group, "tree")
        except KeyError:
            biome_functions.setup_tree_group(mod_group)

    def execute(self, context):
        if self.settings_group:
            settings_group = bpy.data.node_groups[self.settings_group]
            tree_list = settings_group.at.tree_list
        else:
            tree_list, settings_group, t_index, tree = get_tree_vars(self, context)

        if self.action == "ADD":

            self.add_tree(tree_list, settings_group)

            return {"FINISHED"}

        elif self.action == "REMOVE":

            if len(tree_list) == 0:
                return

            control_node = get_node(settings_group, "control")
            collection = get_input(control_node, "collection").default_value

            if tree.selected_tree == "default.png" and len(tree_list) != 1:

                for obj in collection.objects:
                    if "AT_default_obj_" in obj.name:
                        break
                collection.objects.unlink(obj)
            else:
                if tree.selected_tree != "default.png":
                    tree_obj = tree.tree_obj
                    mat = tree_obj.material_slots[0].material
                    for n in mat.node_tree.nodes:
                        if n.type == "TEX_IMAGE" and n.image and n.image.users == 1:
                            bpy.data.images.remove(n.image)
                    bpy.data.materials.remove(mat)
                    bpy.data.meshes.remove(tree_obj.data)

                    if len(tree_list) == 1:
                        obj = helpers.get_default_obj(self, context, control_node)
                        collection.objects.link(obj)

            tree_list.remove(t_index)
            settings_group.at.t_index = min(max(0, t_index - 1), len(tree_list) - 1)

            return {"FINISHED"}

        elif self.action == "UP" or "DOWN":
            up = (self.action == "UP")
            length = len(tree_list) - 1
            new_index = t_index + (-1 if up else 1)

            tree_list.move(new_index, t_index)
            settings_group.at.t_index = max(0, min(new_index, length))

        return {"FINISHED"}


class ALPHATREE_OT_copy_settings_to(bpy.types.Operator):
    """Copy settings from current tree to others"""
    bl_label = ""
    bl_idname = "alpha_tree.copy_settings_to"
    bl_description = "Copy settings from current tree to others"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {"REGISTER", "UNDO"}

    settings_type: bpy.props.EnumProperty(items=(
        ("SCATTERING", "Scattering", "Copy scattering settings"),
        ("WIND", "Wind", "Copy wind settings"),
        ("DISTRIBUTION", "Distribution", "Copy distribution settings"),
        ("FULL_TREES", "Full trees", "Copy full trees to another biome"),
        ("TREE", "Tree settings", "Copy tree settings"),
        ("MATERIAL", "Material", "Copy material settings"),
    ))

    use_groups: bpy.props.BoolProperty()

    groups: bpy.props.StringProperty()

    use_popup: bpy.props.BoolProperty(default=True)

    def filter_inputs(self, inputs):
        """Return the inputs that are part of self.groups"""

        # check if single or multiple groups
        if self.groups and self.groups[0] == "[":
            groups = eval(self.groups)
        else:
            groups = [self.groups]

        filtered = []
        for socket in inputs:
            if not socket.links and not socket.hide_value:
                for group in groups:
                    if socket.name.startswith(group):
                        filtered.append(socket)

        return filtered

    def add_node_copy_items(self, context, node):
        copy_items = context.scene.alpha_trees.copy_items
        filtered_inputs = self.filter_inputs(node.inputs)
        for input in filtered_inputs:
            copy_item = copy_items.add()
            copy_item.name = input.name
            copy_item.hide = input.hide_value
            copy_item.from_node_input = True

    def invoke(self, context, event):

        self.from_node = None
        self.from_data = []
        self.to_nodes = []
        self.to_data = []
        self.to_icon_vals = []
        self.to_draw_groups = []

        at = context.scene.alpha_trees
        copy_items = at.copy_items
        copy_items.clear()
        copy_type = self.settings_type
        biome = get_biome_vars(self, context)[-1]
        bsetting = biome.biome_settings
        tree_list, _, _, from_tree = get_tree_vars(self, context)
        all_bsettings = get_all_bsettings()

        def append_tree_data(self, all_bsettings, mat=False):
            for bsett in all_bsettings:
                for tree in bsett.at.tree_list:
                    if tree != from_tree and tree.selected_tree != "default.png":
                        self.to_data.append(tree)
                        self.to_draw_groups.append(bsett.name.replace(NODE_NAMES["slot"] + "_", ""))
                        self.to_icon_vals.append(tree.icon_value)
                        if mat:
                            self.to_nodes.append(get_mat_control_node(self, context, tree))
                        else:
                            mod_group = helpers.get_mod_group(tree.tree_obj)
                            self.to_nodes.append(get_node(mod_group, "tree"))

        def append_bsettings(self, append_nodes=True):
            for bsett in all_bsettings:
                if bsett != bsetting:
                    self.to_data.append(bsett.at)
                    self.to_draw_groups.append("")
                    self.to_icon_vals.append(get_icon("biome.png"))
                    if append_nodes:
                        self.to_nodes.append(get_node(bsett, "control"))

        # define class attributes (from_node, to_nodes, etc.)
        # non nodes:
        if copy_type == "FULL_TREES":
            self.from_data.extend(tree_list)
            for tree in tree_list:
                copy_item = copy_items.add()
                copy_item.name = tree.name
                copy_item.icon = tree.icon_value
                copy_item.string = tree.selected_tree
            append_bsettings(self, append_nodes=False)

        # node copying:
        elif copy_type in ["SCATTERING", "WIND", "FULL_TREES"]:
            self.from_node = get_node(bsetting, "control")
            append_bsettings(self)
        elif copy_type == "TREE":
            mod_group = helpers.get_mod_group(from_tree.tree_obj)
            self.from_node = get_node(mod_group, "tree")
            append_tree_data(self, all_bsettings)
        elif copy_type == "MATERIAL":
            self.from_node = get_mat_control_node(self, context, from_tree)
            append_tree_data(self, all_bsettings, mat=True)

        def add_copy_item(name, type, value):
            copy_item = copy_items.add()
            copy_item.name = name
            setattr(copy_item, type, value)

        # place at top of list
        if copy_type == "WIND":
            copy_item = copy_items.add()
            copy_item.name = ""
            copy_item.hide = True

        elif copy_type == "TREE":
            add_copy_item("tree_Quality", "string", from_tree.quality)

        elif copy_type == "DISTRIBUTION":
            add_copy_item("Denisty", "string", biome.density.vgroup_enum)
            add_copy_item("Scale", "string", biome.scale.vgroup_enum)
            self.from_data = biome

            for to_biome in get_at_obj(context).at.biome_list:
                if to_biome != biome and to_biome.biome_settings:
                    self.to_data.append(to_biome)
                    self.to_icon_vals.append(0)
                    self.to_draw_groups.append("")

        if self.from_node:
            # add node inputs
            self.add_node_copy_items(context, self.from_node)

        # place at bottom of list
        if copy_type == "SCATTERING":
            add_copy_item("Use scene camera", "bool", bsetting.at.use_scene_cam)

        if self.use_popup:
            return context.window_manager.invoke_props_dialog(self, width=450)
        else:
            return self.execute(context)

    def draw(self, context):
        layout = self.layout
        copy_items = context.scene.alpha_trees.copy_items
        copy_type = self.settings_type

        if copy_type == "DISTRIBUTION":
            layout.label(text="Distribution settings cannot be copied in this blender version :(")
            return

        split = layout.split()
        col = split.column(align=True)
        row = col.row()
        name = self.settings_type.lower().replace("_", " ")
        row.label(text="Copy type: " + name)
        row = row.row(align=True)
        row.active = False
        row.operator(
            "alpha_tree.toggle_copy_items",
            text="",
            emboss=False,
            icon_value=get_icon("toggle.png"),
        ).type = "TOGGLE"
        row.operator(
            "alpha_tree.toggle_copy_items",
            text="",
            emboss=False,
            icon_value=get_icon("invert.png"),
        ).type = "INVERT"

        col2 = col.box().column(align=True)

        for copy_item in copy_items:
            if copy_item.hide:
                if copy_item.name and not copy_item.name.startswith("_"):
                    row = centered_row(col2)
                    row.label(text=copy_item.name, icon="BLANK1")
                else:
                    col2.separator()
            else:
                icon = copy_item.icon if copy_item.icon else 0
                # if self.use_groups and (not self.groups or copy_item.name.split("_")[0] + "_" in eval(self.groups)):
                if len(copy_item.name.split("_")) > 1 and self.use_groups:
                    name = copy_item.name.split("_")[1]
                else:
                    name = copy_item.name
                col2.prop(copy_item, "on", icon_value=icon, text=name, toggle=True)

        col = split.column(align=True)
        row = col.row()
        name = "biome settings" if copy_type in ["SCATTERING", "WIND", "FULL_TREES"] else "trees"
        row.label(text="Copy to: " + name)
        row = row.row(align=True)
        row.active = False
        op = row.operator("alpha_tree.toggle_copy_items", text="", emboss=False, icon_value=get_icon("toggle.png"))
        op.type = "TOGGLE"
        op.data = str(self.to_data)
        op = row.operator("alpha_tree.toggle_copy_items", text="", emboss=False, icon_value=get_icon("invert.png"))
        op.type = "INVERT"
        op.data = str(self.to_data)

        box = col.box().column(align=True)
        last_group = ""
        for data, draw_group, icon_val in zip(
                self.to_data, self.to_draw_groups,
                self.to_icon_vals):  # If I were fancy, I would call the iterator "datum" because it's singular...
            if last_group != draw_group:
                if last_group:
                    box = col.box().column(align=True)
                row = centered_row(box, use_opposite=True, fac=0.05)
                row.active = False
                row.label(text=draw_group)
                last_group = draw_group
            box.prop(data, "copy_to", text=data.name, icon_value=icon_val, toggle=True)

        if len(self.to_data) == 0:
            row = box.row()
            row.alignment = "CENTER"
            row.active = False
            row.label(text=f"No {name} found")

    def copy_node_simple(self, context, to_node):
        """Copy inputs only by name, not accurate for nodes with multiple sockets using the same name"""
        copy_items = context.scene.alpha_trees.copy_items
        from_inputs = self.filter_inputs(self.from_node.inputs)
        matching = True

        for copy_item, from_input in zip(copy_items, from_inputs):
            if copy_item.on and not copy_item.hide:
                try:
                    to_input = to_node.inputs[from_input.name]
                except KeyError:
                    self.report({"ERROR"},
                                f"Nodes do not match, could not copy '{from_input.name}' to '{to_node.id_data.name}'")
                    continue

                to_input.default_value = from_input.default_value
                if helpers.get_list_index(from_input) != helpers.get_list_index(to_input):
                    matching = False
                    continue

        if not matching:
            self.report({"WARNING"}, "Nodes do not match, results may not be as expected")

    def execute(self, context):

        copy_items = context.scene.alpha_trees.copy_items
        copy_type = self.settings_type

        if copy_type == "FULL_TREES":
            bsetting = get_biome_vars(self, context)[-1].biome_settings
            trees = [bsetting.at.tree_list[i.name] for i in copy_items if i.on]
            for to_bsetting in self.to_data:
                if to_bsetting.copy_to:
                    for tree in trees:
                        ALPHATREE_OT_tree_list_actions.add_tree(to_bsetting.tree_list, to_bsetting.id_data)
                        new_tree = to_bsetting.tree_list[-1]
                        new_tree.selected_tree = tree.selected_tree

                        node = get_mat_control_node(self, context, tree)
                        self.from_node = node
                        copy_items.clear()
                        self.add_node_copy_items(context, node)
                        self.copy_node_simple(context, get_mat_control_node(self, context, new_tree))

                        new_tree.quality = tree.quality
                        node = get_node(helpers.get_mod_group(tree.tree_obj), "tree")
                        self.from_node = node
                        copy_items.clear()
                        self.add_node_copy_items(context, node)
                        self.copy_node_simple(context, get_node(helpers.get_mod_group(new_tree.tree_obj), "tree"))

        elif copy_type == "DISTRIBUTION":
            return {"FINISHED"}
            for biome in self.to_data:
                if biome.copy_to:
                    if copy_items[0].on:
                        biome.density.vgroup_enum = self.from_data.density.vgroup_enum
                    if copy_items[1].on:
                        biome.scale.vgroup_enum = self.from_data.scale.vgroup_enum

        # copy data
        if copy_items[0].on:
            if copy_type == "TREE":
                for tree in self.to_data:
                    tree.quality = copy_items[0].string
        if copy_items[-1].on:
            if copy_type == "SCATTERING":
                for bsett_at in self.to_data:
                    bsett_at.use_scene_cam = copy_items[0].bool

        if self.to_nodes:
            # remove non node items
            for i, copy_item in enumerate(copy_items):
                if not copy_item.from_node_input:
                    copy_items.remove(i)

            # copy node values
            for data, node in zip(self.to_data, self.to_nodes):
                if data.copy_to:
                    self.copy_node_simple(context, node)

        on = len([i for i in copy_items if i.on and not i.hide])
        recipients = len([data for data in self.to_data if data.copy_to])
        s1 = "y" if on == 1 else "ies"
        s2 = "" if recipients == 1 else "s"
        self.report({"INFO"}, f"Copied {str(on)} propert{s1} to {str(recipients)} recipient{s2}")

        return {"FINISHED"}


class ALPHATREE_OT_copy_biome_settings(bpy.types.Operator):
    """Make a copy of current biome settings"""
    bl_label = "Copy biome settings"
    bl_idname = "alpha_tree.copy_biome_settings"
    bl_description = "Make a copy of current biome settings"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {"REGISTER", "UNDO"}

    def execute(self, context):
        biome_slot = get_biome_vars(self, context)[-1]
        old_settings = biome_slot.biome_settings
        bpy.ops.alpha_tree.sys_new_settings(add_tree=False)
        new_settings = biome_slot.biome_settings
        new_settings.name = old_settings.name + " Copy"

        for bsetting in get_all_bsettings():
            bsetting.at.copy_to = bsetting == new_settings
        biome_slot.biome_settings = old_settings
        bpy.ops.alpha_tree.copy_settings_to(
            "INVOKE_DEFAULT",
            use_popup=False,
            settings_type='SCATTERING',
            groups="['scatter_', 'cam_', 'wind_']",
        )
        bpy.ops.alpha_tree.copy_settings_to(
            "INVOKE_DEFAULT",
            use_popup=False,
            settings_type='FULL_TREES',
        )

        biome_slot.biome_settings = new_settings
        return {'FINISHED'}


classes = (
    ALPHATREE_OT_update_old_biome,
    ALPHATREE_OT_biome_list_actions,
    ALPHATREE_OT_save_biome,
    ALPHATREE_OT_load_biome,
    ALPHATREE_OT_remove_biome,
    ALPHATREE_OT_manage_biomes,
    ALPHATREE_OT_export_package,
    ALPHATREE_OT_install_package,
    ALPHATREE_OT_new_settings,
    ALPHATREE_OT_remove_settings,
    ALPHATREE_OT_new_vert_group,
    ALPHATREE_OT_remove_vert_group,
    ALPHATREE_OT_edit_vert_group,
    ALPHATREE_OT_tree_list_actions,
    ALPHATREE_OT_copy_settings_to,
    ALPHATREE_OT_copy_biome_settings,
    ALPHATREE_OT_vert_group_actions,
)

# Used for quick testing of the lite verseion
if IS_DEV and VARIANT == "LITE":
    classes = ()


def register():
    for cls in classes:
        bpy.utils.register_class(cls)


def unregister():
    for cls in classes:
        bpy.utils.unregister_class(cls)
