import bpy
from bpy.props import StringProperty, EnumProperty, IntProperty

from ..functions.helper_functions import get_at_obj, get_tree_vars


class ALPHATREE_OT_toggle_viewport(bpy.types.Operator):
    """Togle the visibility of this item in the viewport"""
    bl_label = "Toggle visibility in viewport"
    bl_idname = "alpha_tree.toggle_viewport"
    bl_description = "Shift to toggle,\nCtrl to invert\nCtrl + Shift to isolate"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {"REGISTER", "INTERNAL"}

    index: IntProperty()

    def invoke(self, context, event):
        biome_list = get_at_obj(context).at.biome_list
        on = len([b for b in biome_list if b.show_viewport])

        for biome in biome_list:
            if event.shift and event.ctrl:
                if biome != biome_list[self.index]:
                    biome.show_viewport = False
                else:
                    biome.show_viewport = True
            elif event.shift:
                biome.show_viewport = on < len(biome_list) / 2
            elif event.ctrl:
                biome.show_viewport = not biome.show_viewport
            else:
                biome = biome_list[self.index]
                biome.show_viewport = not biome.show_viewport
                break

        return {"FINISHED"}


class ALPHATREE_OT_toggle_render(bpy.types.Operator):
    """Togle the visibility of this item in the render"""
    bl_label = "Toggle visibility in viewport"
    bl_idname = "alpha_tree.toggle_render"
    bl_description = "Shift to toggle,\nCtrl to invert\nCtrl + Shift to isolate"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {"REGISTER", "INTERNAL"}

    index: IntProperty()

    type: StringProperty()

    def invoke(self, context, event):

        if self.type == "tree":
            list = get_tree_vars(self, context)[0]
        elif self.type == "biome":
            list = get_at_obj(context).at.biome_list
        elif self.type == "obj":
            list = context.scene.alpha_trees.obj_list

        on = len([b for b in list if b.show_render])

        for item in list:
            if event.shift and event.ctrl:
                if item != list[self.index]:
                    item.show_render = False
                else:
                    item.show_render = True
            elif event.shift:
                item.show_render = on < len(list) / 2
            elif event.ctrl:
                item.show_render = not item.show_render
            else:
                item = list[self.index]
                item.show_render = not item.show_render
                break

        return {"FINISHED"}


class ALPHATREE_OT_toggle_copy_items(bpy.types.Operator):
    """Toggle or invert all items"""
    bl_label = "Toggle"
    bl_idname = "alpha_tree.toggle_copy_items"
    bl_description = "Toggle or invert all items"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {"INTERNAL"}

    key: StringProperty()

    type: EnumProperty(
        items=(
            ("TOGGLE", "Toggle", "Turn all items either on or off"),
            ("INVERT", "Invert", "Turn off items on, and on items off"),
        ),
        default="TOGGLE",
    )

    data: StringProperty(default="")

    def execute(self, context):

        keys = self.key.split(",")
        for k in keys:
            if k == "":
                keys.remove(k)

        at = context.scene.alpha_trees
        copy_items = eval(self.data) if self.data else at.copy_items
        toggle = "copy_to" if self.data else "on"

        if self.type == "TOGGLE":
            on = 0
            total = 0
            for item in copy_items:

                search = [k if k.lower() in item.string.lower() else "" for k in keys]
                if all(search):
                    total += 1
                    if getattr(item, toggle):
                        on += 1

        for item in copy_items:
            search = [k if k.lower() in item.string.lower() else "" for k in keys]
            hide = False if self.data else item.hide
            if all(search) and not hide:
                if self.type == "TOGGLE":
                    setattr(item, toggle, True if on < total / 2 else False)
                else:
                    setattr(item, toggle, False if getattr(item, toggle) else True)

        return {"FINISHED"}


classes = (
    ALPHATREE_OT_toggle_viewport,
    ALPHATREE_OT_toggle_render,
    ALPHATREE_OT_toggle_copy_items,
)


def register():
    for cls in classes:
        bpy.utils.register_class(cls)


def unregister():
    for cls in classes:
        bpy.utils.unregister_class(cls)