import bpy

from .constants import AT_LAYER_NAMES, VARIANT
from .info import draw_info
from .icons import get_icon, get_tree_enum_previews, get_tree_previews, reload_all_previews
from .functions import imp_functions, biome_functions
from .functions import helper_functions as helpers
from .functions.imp_functions import draw_open_panel_button, draw_single_material_section
from .functions.helper_functions import draw_tree_search, get_mod_group, get_node, get_input, get_prefs

if VARIANT != "LITE":
    from .operators.biome_ops import (
        ALPHATREE_OT_biome_list_actions,
        ALPHATREE_OT_copy_biome_settings,
        ALPHATREE_OT_new_settings,
        ALPHATREE_OT_remove_settings,
        ALPHATREE_OT_tree_list_actions,
    )
from .operators.gen_ops import ALPHATREE_OT_obj_list_actions

from .weight_layers_addon import WL_functions
from .weight_layers_addon.WL_functions import get_wl_obj as get_at_obj, wrap_text
from .weight_layers_addon.WL_ui import WEIGHTLAYER_PT_main_panel

import os
from os.path import join


class ALPHATREE_UL_GenList(bpy.types.UIList):
    """UIlist for the generator objects"""

    def draw_item(self, context, layout, data, item, icon, active_data, active_propname, index):

        custom_icon_val = get_icon("tree_oak.png")

        custom_icon = 'PARTICLES'

        if self.layout_type in {'DEFAULT', 'COMPACT'}:
            row = layout.row(align=True)
            if item.tree_obj:

                for m in item.tree_obj.material_slots:
                    if m.material:
                        m = m.material
                        if m.at.unsure:
                            row.alert = True
                            break

                row.prop(item.tree_obj, "name", text="", emboss=False, icon_value=custom_icon_val)
            else:
                row.label(text="New Tree " + str(index + 1), icon_value=custom_icon_val)

            op = row.operator("alpha_tree.toggle_render",
                              text="",
                              emboss=False,
                              icon="RESTRICT_RENDER_OFF" if item.show_render else "RESTRICT_RENDER_ON")
            op.index = index
            op.type = "obj"

        elif self.layout_type in {'GRID'}:
            layout.alignment = 'CENTER'
            layout.prop(item, "name", text="", emboss=False, icon=custom_icon)


class ALPHATREE_UL_BiomeList(bpy.types.UIList):
    """UIlist for the particle systems"""

    def draw_item(self, context, layout, data, item, icon, active_data, active_propname, index):

        try:
            custom_icon_val = get_icon("biome.png")
        except KeyError:
            get_tree_enum_previews(self, context)
        custom_icon = 'PARTICLES'

        if self.layout_type in {'DEFAULT', 'COMPACT'}:
            row = layout.row(align=True)

            if item.biome_settings:
                settings_group = item.biome_settings
                row.prop(settings_group.at, "name", icon_value=custom_icon_val, emboss=False, text="")

                op = row.operator("alpha_tree.toggle_viewport",
                                  text="",
                                  emboss=False,
                                  icon="RESTRICT_VIEW_OFF" if item.show_viewport else "RESTRICT_VIEW_ON")
                op.index = index

                op = row.operator("alpha_tree.toggle_render",
                                  text="",
                                  emboss=False,
                                  icon="RESTRICT_RENDER_OFF" if item.show_render else "RESTRICT_RENDER_ON")
                op.index = index
                op.type = "biome"
            else:
                row.label(text="Biome " + str(index), icon_value=custom_icon_val)

        elif self.layout_type in {'GRID'}:
            layout.alignment = 'CENTER'
            layout.prop(item, "name", text="", emboss=False, icon=custom_icon)


def panel_draw_handler():
    # Reloading the icons requires changing blender properties, which isn't allowed in UI draw functions,
    # so use a draw handler instead.
    global to_reload
    global reload_all
    if not reload_all:
        for tree in to_reload:
            tree_name = tree.selected_tree if tree.selected_tree else tree.previous_tree
            get_tree_previews(data=tree, high_res_tree=tree_name, reload=True)
            if not tree.selected_tree:
                tree.selected_tree = tree.previous_tree
        to_reload.clear()
    else:
        reload_all_previews(None, bpy.context)
        reload_all = False


to_reload = []
reload_all = False
reloaded = False  # Used to keep track of whether the preview cache has been reloaded.


class ALPHATREE_UL_TreeList(bpy.types.UIList):
    """UIlist for the trees"""

    def draw_item(self, context, layout, data, item, icon, active_data, active_propname, index):

        try:
            custom_icon_val = get_icon("biome.png")
        except KeyError:
            get_tree_enum_previews(self, context)

        global reloaded
        global to_reload
        # check if tree in current library
        try:
            custom_icon_val = get_icon(item.selected_tree, data=item, type="preview")
        except KeyError:
            # Try updating the tree preview cache as a last attempt. Only do this once so the UI doesn't slow down.
            # Uses a global variable to keep track of whether it has been reloaded,
            # which is reset when drawing returns to normal.
            if not reloaded:
                # reload the previews in the draw handler.
                to_reload.append(item)
            reloaded = True
            custom_icon_val = 1
            row = layout.row(align=True)
            row.alert = True
            row.label(text=f"Tree {item.selected_tree} not in current library", icon="ERROR")
            row.alert = False

            draw_info(row, "Tree not found", "not_current_lib")
            return
        reloaded = False

        if self.layout_type in {'DEFAULT', 'COMPACT'}:
            row = layout.row(align=True)
            if custom_icon_val != 1 and item.selected_tree != "default.png":
                # custom_icon_val = item.tree_obj.active_material.preview.icon_id
                custom_icon_val = item.icon_value

            row.prop(item, "name", text="", emboss=False, icon_value=custom_icon_val)

            tree_control_node = get_node(get_mod_group(item.tree_obj), "tree")
            show_render = get_input(tree_control_node, "tree_show_render").default_value
            op = row.operator(
                "alpha_tree.toggle_render",
                text="",
                emboss=False,
                icon="RESTRICT_RENDER_OFF" if show_render else "RESTRICT_RENDER_ON",
            )
            op.index = index
            op.type = "tree"
            # ^ love this, can't believe I didn't think of this for so long lol

        elif self.layout_type in {'GRID'}:
            layout.alignment = 'CENTER'
            layout.prop(item, "name", text="", emboss=False, icon_value=custom_icon_val)


# Menus
class ALPHATREE_MT_BiomeMenu(bpy.types.Menu):
    bl_label = "Biome menu"
    bl_idname = "ALPHATREE_MT_BiomeMenu"

    def draw(self, context):
        layout = self.layout
        layout.row().operator("alpha_tree.save_biome", icon="FILE_TICK")
        layout.row().operator("alpha_tree.load_biome", icon="SORT_ASC")
        layout.row().operator("alpha_tree.manage_saved_biomes", icon="GREASEPENCIL")

        draw_info(layout, "Save/Load", "biome_menu_ops", text="Info")


class ALPHATREE_MT_BatchMenu(bpy.types.Menu):
    bl_label = "Batch menu"
    bl_idname = "ALPHATREE_MT_BatchMenu"

    def draw(self, context):
        layout = self.layout
        layout.row().operator("alpha_tree.auto_mat_type", text="Autoname all mat types",
                              icon="OUTLINER_DATA_FONT").all = True
        layout.row().operator("alpha_tree.add_objs_to_list", text="Add selection objects",
                              icon="OBJECT_DATA").type = "SELECTION"
        layout.row().operator("alpha_tree.add_objs_to_list", text="Add collection objects",
                              icon="GROUP").type = "COLLECTION"


class ALPHATREE_MT_TreeMenu(bpy.types.Menu):
    bl_label = "Tree menu"
    bl_idname = "ALPHATREE_MT_TreeMenu"

    def draw(self, context):
        layout = self.layout
        layout.row().operator("alpha_tree.multi_import", text="Multi add",
                              icon_value=get_icon("multi_import.png")).particle = True
        row = layout.row()
        row.enabled = len(helpers.get_tree_vars(self, context)[0]) != 0
        op = row.operator("alpha_tree.copy_settings_to", text="Copy trees", icon="COPYDOWN")
        op.settings_type = 'FULL_TREES'
        op.use_groups = False  # draw all sockets

        draw_info(layout, "Multi add/Copy trees", "tree_menu_ops", text="Info", emboss=True)


# Panels


class ALPHATREE_PT_gen_panel(bpy.types.Panel):
    """Generator panel"""
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_label = "Alpha creator"
    bl_category = "Alpha Trees"
    bl_options = {"DEFAULT_CLOSED"}

    @classmethod
    def poll(cls, context):
        return get_prefs().lib_path != ""

    def draw(self, context):
        prefs = get_prefs()
        at = context.scene.alpha_trees
        layout = self.layout

        # check for botaniq
        if prefs.show_botaniq_operator:
            for addon in context.preferences.addons:
                if "botaniq" in addon.module:
                    botaniq_col = layout.box().column()
                    botaniq_col.scale_y = 1.4

                    if prefs.botaniq_version == "OLD":
                        op = botaniq_col.operator("material.biq_spawn_asset",
                                                  icon_value=get_icon("tree_oak.png"),
                                                  text="Plant botaniq tree")
                        if not op:
                            botaniq_col.scale_y = 1
                            botaniq_col.label(text="Botaniq version selected in preferences")
                            botaniq_col.label(text="is not correct:")
                            botaniq_col.prop(prefs, "botaniq_version")
                    else:
                        op = botaniq_col.operator("botaniq.spawn_asset",
                                                  icon_value=get_icon("tree_oak.png"),
                                                  text="Plant botaniq tree")
                        if not op:
                            botaniq_col.scale_y = 1
                            botaniq_col.label(text="Botaniq version selected in preferences")
                            botaniq_col.label(text="is not correct:")
                            botaniq_col.prop(prefs, "botaniq_version")

        column = layout.column(align=True)
        col = column.box().column(align=True)
        row = col.row()
        row.prop(at, "generation_type", expand=True)
        col = column.box().column(align=False)
        tree_obj = None
        obj_list = at.obj_list
        if at.generation_type == "SINGLE":
            col.label(text="Tree object:")
            col.prop(at, "tree_obj", text="")
            tree_obj = at.tree_obj
        else:

            WL_functions.draw_ui_list(col,
                                      list=obj_list,
                                      list_class=ALPHATREE_UL_GenList,
                                      actions_operator=ALPHATREE_OT_obj_list_actions,
                                      data=at,
                                      list_setting="obj_list",
                                      list_name="The_gen_list",
                                      index_data=at,
                                      index_name="obj_index",
                                      menu="ALPHATREE_MT_BatchMenu")

            if obj_list:
                item = obj_list[at.obj_index]

                col.prop(item, "tree_obj", text="")
                tree_obj = item.tree_obj

        if tree_obj:
            column = layout.column(align=True)
            if tree_obj.type == "EMPTY":
                row = column.box().column(align=False).row(align=True)
                row.alert = True
                row.scale_y = 1.4
                row.scale_x = 1.2
                row.operator("alpha_tree.make_editable", text="Make tree local", icon="MESH_DATA")
                draw_info(row, "Make tree local", "make_editable", emboss=True)

            else:
                col = column.box().column(align=False)
                mat_row = col.row(align=True)
                if len(tree_obj.material_slots) > 1:
                    mat_row.scale_y = 1.1
                    mat_row.scale_x = 1.07
                    mat_row.prop(at, "show_mat_settings", text="", icon="MATERIAL_DATA")
                    lab_row = mat_row.row()

                    for mat in tree_obj.material_slots:
                        if mat.material:
                            mat = mat.material
                            if mat.at.unsure:
                                lab_row.alert = True
                                break

                    lab_row.label(text="Material types")
                    mat_row.operator("alpha_tree.auto_mat_type", text="", icon="OUTLINER_DATA_FONT", emboss=False)
                    draw_info(mat_row, "Material types", "mat_types")

                else:
                    mat_row.alert = True
                    mat_row.label(text="Missing materials")
                    draw_info(mat_row, "Missing materials", "missing_mats")

            if len(tree_obj.material_slots) > 1 and at.show_mat_settings:
                col = column.box().column(align=False)
                unsure_row = col.row()
                unsure_row.alert = True
                unsure = False
                for mat in tree_obj.material_slots:
                    if mat.material:
                        mat = mat.material
                        row = col.row()
                        if mat.at.unsure:
                            row.alert = True
                            if not unsure:
                                unsure_row.label(text="Unsure about some materials")
                                unsure = True
                        row.label(text=mat.name + ":", icon_value=mat.preview.icon_id)
                        row.alert = False
                        row.prop(mat.at, "type", text="")

        # settings
        column = layout.column(align=True)
        col = column.box().column()
        row = col.row(align=True)
        row.scale_y = 1.1
        row.scale_x = 1.07
        row.prop(at, "show_gen_settings", text="", icon="PREFERENCES")
        row.label(text="Settings")

        draw_info(row, "Render settings", "gen_settings")

        if at.show_gen_settings:
            col2 = column.box().column(align=True)
            # col2 = col.column(align=True)
            col2.scale_y = 1.1
            col2.label(text="General:")
            col2.prop(at, "border_padding", text="Border padding")

            col2.label(text="Rendering:")
            col2.prop(at, "resolution", text="Resolution")
            col2.prop(context.scene.cycles, "samples", text="Samples (50 max)")

            col2.label(text="Library path:")
            col2.prop(prefs, "lib_path")

            if prefs.show_dev_options:
                col2.label(text="Debug:")
                col2.prop(at, "diff_render", text="Render diffuse", toggle=True)
                col2.prop(at, "nor_render", text="Render normal", toggle=True)
                col2.prop(at, "mask_render", text="Render mask", toggle=True)
                col2.prop(at, "preview_render", text="Render preview", toggle=True)
                col2.prop(at, "remove_extra_masks", text="Remove extra masks")

        column = layout.column(align=True)
        err_row = column.row()
        row = column.box().row(align=True)
        row.scale_y = 1.57
        row.scale_x = 1.26
        if at.generation_type == "BATCH":
            if not obj_list:
                row.active = False
        if tree_obj:
            if tree_obj.type != "EMPTY":
                if len(tree_obj.material_slots) > 1:
                    row.active = True
                else:
                    row.active = False
            else:
                row.active = False
        else:
            row.active = False

        if tree_obj:
            if len(tree_obj.particle_systems) > 0:
                row.active = False
                err_row.alert = True
                err_row.label(text="Please convert particles to a single mesh")
                draw_info(err_row, "Particles", "Particle error")

        row.operator("alpha_tree.execute_all", text="Bake maps", icon="RESTRICT_RENDER_OFF")
        row.operator("alpha_tree.open_folder", text="", icon="FILE_FOLDER")


class ALPHATREE_PT_import_panel(bpy.types.Panel):
    """Biome panel"""
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_label = "Import"
    bl_category = "Alpha Trees"

    def draw_header_preset(self, context):
        if VARIANT != "LITE":
            WL_functions.draw_headers(self, context)

    def draw(self, context):
        layout = self.layout
        at = bpy.context.scene.alpha_trees
        prefs = get_prefs()

        message = prefs.check_lib_path()
        if message:
            box = layout.box()
            col = box.column(align=True)
            col.scale_y = 0.9
            wrap_text(context, message, col, True)
            box.prop(prefs, "lib_path")
            return

        obj = get_at_obj(context)
        if obj is not None and VARIANT != "LITE":
            row = layout.row(align=True)
            row.prop(
                at,
                "import_type",
                expand=True,
            )

        if at.import_type == "PARTICLE" and obj is not None:

            if VARIANT == "LITE":
                box = layout.box()
                col = box.column(align=True)
                col.scale_y = 0.9
                # You should only see this if you've tampered with the files, or I've made a big mistake...
                # Either way, good for you.
                message = "If you want a free upgrade to Alpha Trees pro,\
                        send a picture of a piece of paper with your email written on it,\
                        and below that, your favourite joke to upgradepls@strikedigital.xyz :)".replace("  ", "")

                wrap_text(context, message, col, True)
                return

            biome_list, _, _, _, biome = helpers.get_biome_vars(self, context)

            if not biome:
                for psystem in obj.particle_systems:
                    if psystem.settings:
                        if "AT_BIOME_" in psystem.settings.name:
                            box = layout.box()
                            col = box.column(align=True)
                            col.scale_y = 0.9
                            text = "It looks like this object has biome settings from an older version of Alpha Trees,\
                                do you want to update them?".replace("  ", "")
                            wrap_text(context, text, col, centered=True)
                            col = col.column(align=True)
                            col.separator(factor=0.75)
                            col.scale_y = 1.2
                            col.operator("alpha_tree.update_old_biomes")
                            break

            col = WL_functions.draw_ui_list(
                layout,
                list=biome_list,
                list_class=ALPHATREE_UL_BiomeList,
                actions_operator=ALPHATREE_OT_biome_list_actions,
                data=obj.at,
                list_setting="biome_list",
                list_name="The_list",
                index_data=obj.at,
                index_name="b_index",
                menu="ALPHATREE_MT_BiomeMenu",
                rows=3 if len(biome_list) < 2 else 5,
            )

            col = col.column()
            draw_info(col, "Biomes", "biomes")

            # settings row
            if biome_list:

                row = layout.row(align=True)
                row.prop(biome, "biome_settings_enum", text="", icon_only=True, icon="SETTINGS")

                if not biome.biome_settings:
                    row.operator(ALPHATREE_OT_new_settings.bl_idname, text="New settings", icon="ADD")
                    return

                settings_group = biome.biome_settings
                row.prop(settings_group.at, "name", text="")
                row.operator(ALPHATREE_OT_new_settings.bl_idname, text="", icon="ADD")
                sub = row.row(align=True)
                sub.scale_x = 0.325
                sub.operator(
                    ALPHATREE_OT_copy_biome_settings.bl_idname,
                    text=str(biome.biome_settings.users - 1),
                )
                row.operator(ALPHATREE_OT_remove_settings.bl_idname, text="", icon="REMOVE")

        else:
            draw_tree_search(layout, at)

            # preview enum and operators
            row = layout.row(align=True)
            col = row.column(align=True)
            cc = col.column(align=True)
            scale_y = 6

            cc.scale_y = 1
            cc.operator("alpha_tree.select_random_tree", text="", icon_value=get_icon("random.png"))

            cc = col.column(align=True)
            cc.scale_y = scale_y - 1
            cc.scale_x = 0.92
            cc.operator("alpha_tree.change_tree", text="", icon="TRIA_LEFT").forward = False
            cc = col.column(align=True)
            cc.scale_y = 1
            cc.operator("alpha_tree.reload_previews", text="", icon="FILE_REFRESH")

            # // DONE: Can't show preview labels until I find a way to not make them freak out...
            # https://blenderartists.org/t/what-causes-enumproperty-to-do-this-ui-bug/1237713/7?u=strike_digital
            row.template_icon_view(at, "selected_tree", scale=scale_y + 1, scale_popup=6, show_labels=True)

            col = row.column(align=True)

            cc = col.column(align=True)
            cc.scale_y = 1
            cc.operator("alpha_tree.edit_tree_settings", text="", icon="GREASEPENCIL")

            cc = col.column(align=True)
            cc.scale_y = scale_y - 1
            cc.scale_x = 0.92
            cc.operator("alpha_tree.change_tree", text="", icon="TRIA_RIGHT").forward = True
            cc = col.column(align=True)
            cc.scale_y = 1
            cc.operator("alpha_tree.multi_import", text="", icon_value=get_icon("multi_import.png"))

            # import button + settings
            if at.selected_tree != "default.png":

                col = layout.column(align=True)
                row = col.row(align=True)
                row.scale_y = 1.8
                # row.scale_x = 1.3
                row.operator("alpha_tree.import_tree",
                             icon_value=get_icon("tree3.png"),
                             text="Import:  " + at.selected_tree.split("_")[0])
                row.prop(at, "show_import_settings", text="", icon="PREFERENCES")

                if at.show_import_settings:
                    diff_name = at.selected_tree.replace("preview.png", "diff-leaf.jpg")
                    box = col.box()
                    if os.path.isfile(join(prefs.lib_path, "rendered_maps", "low", diff_name)):
                        row = box.row(align=True)
                        row.prop(at, "quality", text="")
                    else:
                        row = box.row(align=True)
                        row.operator("alpha_tree.generate_lods").tree_name = at.selected_tree
                        draw_info(row, "Create LODs", "create_lods")

            if not obj:
                return

            control_node = helpers.get_mat_control_node(self, context, obj)

            # TRACKING
            if obj.at.is_at:
                layout.separator()
                col = layout.column(align=True)
                boxcol = col.box().column(align=True)
                row = boxcol.row(align=True)

                draw_open_panel_button(context, row, at, at.show_track_settings, "show_track_settings", "CON_TRACKTO")
                row.label(text="Tracking")

                draw_info(row, "Tracking settings", "tracking_settings")

                row.operator(
                    "alpha_tree.copy_settings_to_single",
                    text="",
                    icon="COPYDOWN",
                    emboss=False,
                ).type = "TRACKING"

                if at.show_track_settings:
                    boxcol = col.box().column(align=False)
                    row = boxcol.row(align=True)
                    mod_group = [m for m in obj.modifiers if m.type == "NODES"][0].node_group
                    tree_control_node = get_node(mod_group, "single_tree")

                    obj_at = obj.at
                    row2 = row.row(align=True)
                    row2.enabled = not obj_at.use_scene_cam
                    row2.prop(
                        get_input(tree_control_node, "single_tree_cam"),
                        "default_value",
                        text="",
                        icon="OUTLINER_DATA_CAMERA",
                    )
                    row.prop(obj_at, "use_scene_cam", text="", icon="SCENE_DATA")

                    if obj_at.camera:
                        boxcol.separator(factor=0.2)
                        row = boxcol.row(align=True)
                        row.prop(obj_at, "rotation_offset", icon="UV_SYNC_SELECT")
                        row.prop(obj_at, "invert_track", text="", icon="UV_SYNC_SELECT")

            # MATERIAL
            # This gets it's own functions since it is also drawn in the Scatter panel.
            draw_single_material_section(context, layout, obj)


class ALPHATREE_PT_settings_panel(bpy.types.Panel):
    """Tree panel"""
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_label = "Settings"
    bl_category = "Alpha Trees"

    @classmethod
    def poll(cls, context):

        if VARIANT == "LITE":
            return False

        if not get_prefs().lib_path:
            return False

        if not get_at_obj(context):
            return False

        at = context.scene.alpha_trees
        obj = get_at_obj(context)
        biome_list = obj.at.biome_list

        if not biome_list:
            return False

        return at.import_type == "PARTICLE" and biome_list

    def draw(self, context):
        layout = self.layout
        at = context.scene.alpha_trees
        _, _, _, slot_node, biome = helpers.get_biome_vars(self, context)

        if not biome.biome_settings:
            row = layout.row()
            socket = slot_node.inputs["Density Max"]
            row.alert = socket.default_value == 0
            socket.draw(context, row, slot_node, "Density")
            return

        control_node = get_node(slot_node, "control")

        # SCATTERING SETTINGS
        col = layout.column(align=True)
        row = col.box().row(align=True)

        draw_open_panel_button(context, row, at, at.show_scatter_settings, "show_scatter_settings", "PARTICLES")
        row.label(text="Scattering")

        row2 = row.column(align=False)
        draw_info(row2, "Scatter settings", "scatter_settings")

        op = row.operator("alpha_tree.copy_settings_to", text="", icon="COPYDOWN", emboss=False)
        op.groups = "['scatter_', 'cam_']"
        op.settings_type = "SCATTERING"

        if at.show_scatter_settings:

            box = col.box()
            col = box.column(align=True)
            get_input(slot_node, "seed").draw(context, col, slot_node, "Seed")
            col = box.column(align=True)

            alert = [get_input(control_node, "density"),
                     get_input(control_node, "scale")]  # these inputs will be red when 0
            biome_functions.draw_input_group(context, col, control_node, "scatter", alert=alert)

            # camera
            col = box.column(align=True)
            row = col.row(align=True)
            row1 = row.row(align=True)
            cam_input = get_input(control_node, "camera")
            row1.prop(cam_input, "default_value", text="", icon="CAMERA_DATA")
            row1.enabled = not slot_node.node_tree.at.use_scene_cam
            row = row.row(align=True)
            row.prop(slot_node.node_tree.at, "use_scene_cam", text="", icon="SCENE_DATA")

            if cam_input.default_value:
                row = col.row(align=True)
                row.prop(
                    get_input(control_node, "padding"),
                    "default_value",
                    text="Padding",
                )

        # WIND
        col = layout.column(align=True)
        boxcol = col.box().column()
        row = boxcol.row(align=True)

        draw_open_panel_button(context, row, at, at.show_wind_settings, "show_wind_settings", "FORCE_WIND")
        row.label(text="Wind")

        draw_info(row, "Wind settings", "wind_settings")

        op = row.operator("alpha_tree.copy_settings_to", text="", icon="COPYDOWN", emboss=False)
        op.settings_type = "WIND"
        op.groups = "['wind_']"

        if at.show_wind_settings:
            boxcol = col.box().column(align=True)
            col = boxcol.box().column(align=True)
            biome_functions.draw_input_group(
                context,
                col,
                control_node,
                "wind_",
            )

        # DISTRIBUTION

        col = layout.column(align=True)
        boxcol = col.box().column()
        row = boxcol.row(align=True)

        draw_open_panel_button(context, row, at, at.show_weight_settings, "show_weight_settings", "WPAINT_HLT")
        row.label(text="Distribution")

        row2 = row.column(align=False)
        draw_info(row2, "Distribution settings", "dist_settings")

        op = row.operator("alpha_tree.copy_settings_to", text="", icon="COPYDOWN", emboss=False)
        op.settings_type = "DISTRIBUTION"
        op.groups = ""  # draw all sockets

        if at.show_weight_settings:
            boxcol = col.box().column()

            def draw_distribution(density):
                dist = biome.density if density else biome.scale
                obj = dist.id_data
                try:
                    vgroup = dist.vgroup
                except AttributeError:
                    vgroup = None

                if at.show_vgroup_selector:
                    row = boxcol.row(align=True)
                    row.prop(dist, "vgroup_enum", text="", icon="MESH_DATA", icon_only=True)

                if vgroup:

                    if at.show_vgroup_selector:
                        row.prop(dist, "vgroup_name", text="")

                        op = row.operator("alpha_tree.vert_group_actions", text="", icon="ADD")
                        op.density = density
                        op.action = "ADD"

                        op = row.operator("alpha_tree.vert_group_actions", text="", icon="REMOVE")
                        op.density = density
                        op.action = "REMOVE"

                        boxcol.separator(factor=0.25)
                        row = boxcol.row(align=True)
                        depress = obj.vertex_groups.active == dist.vgroup and obj.mode == "WEIGHT_PAINT"
                        op = row.operator("alpha_tree.vert_group_actions",
                                          text="Edit",
                                          icon="BRUSH_DATA",
                                          depress=depress)
                        op.density = density
                        op.action = "EDIT"

                        socket = get_input(slot_node, "use_density_attr" if density else "use_scale_attr")
                        row.prop(socket, "default_value", text="Invert", icon="ARROW_LEFTRIGHT")
                    # frozen = slot.manual_refresh
                    # boxcol.prop(
                    #     slot,
                    #     "manual_refresh",
                    #     text="Unfreeze" if frozen else "Freeze",
                    #     toggle=True,
                    #     icon="FREEZE",
                    # )

                    slot = obj.weight_layers.layer_stacks[dist.stack_slot_name]
                    if slot.stack_group:
                        boxcol.separator()
                        row = boxcol.row(align=True)
                        row.prop(
                            at,
                            "show_weight_layers",
                            text="",
                            emboss=False,
                            icon="DISCLOSURE_TRI_DOWN" if at.show_weight_layers else "DISCLOSURE_TRI_RIGHT",
                        )
                        row.label(text="Weight layers")
                        if at.show_weight_layers:
                            idx = list(obj.weight_layers.layer_stacks).index(slot)
                            slot.stack_group.wl.draw_layer_stack(
                                context,
                                boxcol,
                                stack_idx=idx,
                                exclude=AT_LAYER_NAMES.values(),
                            )

                else:
                    op = row.operator("alpha_tree.vert_group_actions", text="New vertex group", icon="ADD")
                    op.density = density
                    op.action = "ADD"

                    socket = get_input(slot_node, "use_density_attr" if density else "use_scale_attr")
                    row.prop(socket, "default_value", text="", icon="ARROW_LEFTRIGHT")

            # density

            row = boxcol.row(align=True)
            row.prop(biome, "distribution_type", expand=True)
            boxcol.separator()
            row = boxcol.row(align=True)
            row.prop(
                at,
                "show_vgroup_selector",
                text="",
                emboss=False,
                icon="DISCLOSURE_TRI_DOWN" if at.show_vgroup_selector else "DISCLOSURE_TRI_RIGHT",
            )
            row.label(text="Vertex group")
            if biome.distribution_type == "DENSITY":
                draw_distribution(density=True)
            else:
                draw_distribution(density=False)


class ALPHATREE_PT_tree_panel(bpy.types.Panel):
    """Tree panel"""
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_label = "Trees"
    bl_category = "Alpha Trees"

    @classmethod
    def poll(cls, context):
        if not ALPHATREE_PT_settings_panel.poll(context):
            return False
        biome = helpers.get_biome_vars(None, context)[-1]
        if not biome.biome_settings:
            return False
        return True

    def draw(self, context):
        layout = self.layout
        tree_list, settings_group, _, tree = helpers.get_tree_vars(self, context)
        prefs = get_prefs()
        at = context.scene.alpha_trees

        col = WL_functions.draw_ui_list(
            layout,
            list=tree_list,
            list_class=ALPHATREE_UL_TreeList,
            actions_operator=ALPHATREE_OT_tree_list_actions,
            data=settings_group.at,
            list_setting="tree_list",
            list_name="The_other_list",
            index_data=settings_group.at,
            index_name="t_index",
            menu="ALPHATREE_MT_TreeMenu",
            rows=3 if len(tree_list) < 2 else 4,
        )

        if tree_list:

            try:
                _ = get_icon(tree.selected_tree, data=tree, type="preview")
            except KeyError:
                box = layout.box()
                row = box.row(align=True)
                row.alert = True
                row.label(text=f"Tree{tree.selected_tree} is not in current library:")
                row.alert = False

                draw_info(row, "Tree not found", "not_current_lib")
                box.prop(prefs, "lib_path")
                box.template_icon_view(tree, "selected_tree", show_labels=True, scale_popup=6)
                draw_tree_search(box, tree)
                return

            col = layout.box().column(align=True)
            col.template_icon_view(tree, "selected_tree", show_labels=True, scale_popup=6)

            draw_tree_search(layout, tree)

            if tree.selected_tree != "default.png":

                # TREE SETTINGS
                col = layout.column(align=True)
                box = col.box()
                row = box.row(align=True)
                draw_open_panel_button(
                    context,
                    row,
                    at,
                    at.show_tree_settings,
                    "show_tree_settings",
                    get_icon("tree3.png"),
                )
                row.label(text="Tree")

                row2 = row.row()
                draw_info(row2, "Tree settings", "tree_settings")

                op = row.operator("alpha_tree.copy_settings_to", text="", icon="COPYDOWN", emboss=False)
                op.settings_type = "TREE"
                op.use_groups = True

                if at.show_tree_settings:
                    box = col.box()
                    col = box.column(align=True)

                    # quality
                    diff_name = tree.selected_tree.replace("preview.png", "diff-leaf.jpg")
                    row = col.row(align=True)
                    if os.path.isfile(join(prefs.lib_path, "rendered_maps", "low", diff_name)):
                        row.prop(tree, "quality", text="")
                    else:
                        row.operator("alpha_tree.generate_lods").tree_name = tree.selected_tree
                        draw_info(row, "Create LODs", "create_lods")

                    mod_group = helpers.get_mod_group(tree.tree_obj)
                    tree_node = get_node(mod_group, "tree")

                    biome_functions.draw_input_group(
                        context,
                        col,
                        tree_node,
                        "tree_",
                    )

                # MATERIALS
                col = layout.column(align=True)
                boxcol = col.box().column(align=True)
                row = boxcol.row(align=True)

                draw_open_panel_button(
                    context,
                    row,
                    at,
                    at.show_material_settings,
                    "show_material_settings",
                    "MATERIAL",
                )
                row.label(text="Material")

                row2 = row.row()
                draw_info(row2, "Material settings", "material_settings")

                op = row.operator("alpha_tree.copy_settings_to", text="", icon="COPYDOWN", emboss=False)
                op.settings_type = "MATERIAL"
                op.use_groups = True

                if at.show_material_settings:
                    boxcol = col.column(align=True)

                    object = tree.tree_obj
                    control_node = get_node(object.material_slots[0].material.node_tree, "mat_control")
                    imp_functions.draw_material_settings(context, boxcol, control_node)


class ALPHATREE_PT_package_panel(bpy.types.Panel):
    """Package creator panel"""
    bl_space_type = "VIEW_3D"
    bl_context = "objectmode"
    bl_region_type = "UI"
    bl_label = "Package creator"
    bl_category = "Alpha Trees"
    bl_options = {"DEFAULT_CLOSED"}

    @classmethod
    def poll(cls, context):
        return get_prefs().show_package_creator

    def draw(self, context):
        prefs = get_prefs()
        layout = self.layout
        box = layout.box()
        row = box.row()
        row.label(text="Export directory:")
        draw_info(row, "Packages", "package_creator")

        box.prop(prefs, "export_dir", text="")

        if os.path.exists(bpy.path.abspath(prefs.export_dir)):
            row = box.row(align=True)
            row.scale_y = 1.5
            row.operator("alpha_tree.export_package", icon="EXPORT")
        else:
            row = box.row()
            row.alert = True
            row.label(text="Not a valid filepath")

            row = box.row(align=True)
            row.scale_y = 1.5
            row.operator("alpha_tree.export_package", icon="EXPORT")
            row.enabled = False

        box = layout.box()
        box.scale_y = 1.5
        box.operator("alpha_tree.install_package", icon="IMPORT")


# Scatter5 Panel
class ALPHATREE_UL_ScatterTreeList(bpy.types.UIList):
    """UIlist for trees created with scatter"""

    def draw_item(self, context, layout: bpy.types.UILayout, data, item, icon, active_data, active_propname, index):

        if self.layout_type in {'DEFAULT', 'COMPACT'}:
            row = layout.row(align=True)

            material = item.material_slots[0].material
            icon_id = material.at.icon_value
            name = item.name.split("_")[0]

            try:
                row.label(text=name, icon_value=icon_id)
            except ValueError:
                row.label(text=name, icon_value=get_icon("tree3"))

    def filter_items(self, context, data, property):

        flt_flags = []
        flt_neworder = []
        collection = getattr(data, property)

        flt_flags = [self.bitflag_filter_item] * len(collection)

        for i, obj in enumerate(collection):
            if not obj.at.is_at:
                # For some reason we have to use weird bit operatons in order to filter the items, god knows why...
                # Check the UIList documentation for more info.
                # & is bitwise and, and ~ is bitwise not, aka set that item to False
                flt_flags[i] &= ~self.bitflag_filter_item

        return flt_flags, flt_neworder


class ALPHATREE_PT_scatter_material_panel(bpy.types.Panel):
    """Shows the materials of Alpha Trees that have been created with Scatter5"""
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_label = "Alpha Trees"
    bl_category = "Scatter5"

    @classmethod
    def get_scatter_collection(cls, context):
        obj = context.scene.scatter5.emitter
        if not obj:
            return None
        sc = obj.scatter5
        system = sc.get_psy_active()

        if not system:
            return None

        collection = system.s_instances_coll_ptr
        return collection

    @classmethod
    def get_at_scatter_objs(cls, context):
        collection = cls.get_scatter_collection(context)
        if not collection:
            return []
        objects = collection.objects
        at_objs = []
        for obj in objects:
            if obj.at.is_at:
                at_objs.append(obj)

        return at_objs

    @classmethod
    def poll(cls, context):
        return cls.get_at_scatter_objs(context)

    def draw(self, context):
        collection = self.get_scatter_collection(context)
        objects = self.get_at_scatter_objs(context)

        layout: bpy.types.UILayout = self.layout

        list_cls = ALPHATREE_UL_ScatterTreeList
        layout.template_list(
            list_cls.bl_rna.identifier,
            "",
            collection,
            "objects",
            collection,
            "at_scatter_obj_list_active_idx",
            rows=len(objects),
        )

        try:
            obj = objects[collection.at_scatter_obj_list_active_idx]
        except IndexError:
            return

        try:
            draw_single_material_section(context, layout, obj)
            # control_node = helpers.get_mat_control_node(self, context, obj)
            # draw_material_settings(context, layout, control_node)
        except (AttributeError, KeyError):
            return


classes = [
    ALPHATREE_UL_GenList,
    ALPHATREE_MT_BiomeMenu,
    ALPHATREE_MT_BatchMenu,
    ALPHATREE_MT_TreeMenu,
    ALPHATREE_PT_gen_panel,
    ALPHATREE_PT_import_panel,
    ALPHATREE_PT_settings_panel,
    WEIGHTLAYER_PT_main_panel,
    ALPHATREE_PT_tree_panel,
    ALPHATREE_PT_package_panel,
    ALPHATREE_UL_BiomeList,
    ALPHATREE_UL_TreeList,
    # On talking to Dorian (BD3D), this may not be a good thing to include in a major release.
    # Feel free to uncomment these lines though to be able to edit AT material settings in the scatter5 panel
    # ALPHATREE_UL_ScatterTreeList,
    # ALPHATREE_PT_scatter_material_panel,
]
draw_handler = None


def register():
    for cls in classes:
        bpy.utils.register_class(cls)

    global reload_all
    reload_all = True

    global draw_handler
    draw_handler = bpy.types.SpaceView3D.draw_handler_add(panel_draw_handler, (), "UI", "POST_PIXEL")


def unregister():
    global draw_handler
    bpy.types.SpaceView3D.draw_handler_remove(draw_handler, "UI")

    for cls in classes:
        bpy.utils.unregister_class(cls)
